=== modified file 'HACKING'
--- HACKING	2006-10-11 01:17:15 +0000
+++ HACKING	2006-10-13 07:22:23 +0000
@@ -209,6 +209,15 @@
 > factory, then yes, foo_factory is what I would use.
 
 
+Registries
+----------
+
+Several places in Bazaar use (or will use) a registry, which is a 
+mapping from names to objects or classes.  The registry allows for 
+loading in registered code only when it's needed, and keeping
+associated information such as a help string or description.
+
+
 Lazy Imports
 ------------
 
@@ -311,6 +320,7 @@
 
 Writing tests
 =============
+
 In general tests should be placed in a file named test_FOO.py where 
 FOO is the logical thing under test. That file should be placed in the
 tests subdirectory under the package being tested.
@@ -345,6 +355,11 @@
     command changes its name or signature. Ideally only the tests for a
     given command are affected when a given command is changed.
 
+ 4. If you have a test which does actually require running bzr in a
+    subprocess you can use ``run_bzr_subprocess``. By default the spawned
+    process will not load plugins unless ``--allow-plugins`` is supplied.
+
+
 Doctests
 --------
 

=== modified file 'NEWS'
--- NEWS	2006-10-13 08:53:34 +0000
+++ NEWS	2006-10-13 10:50:33 +0000
@@ -1,13 +1,28 @@
 IN DEVELOPMENT
 
   IMPROVEMENTS:
+
+    * urllib implements keep-alive connections and share them
+      (Vincent Ladeuil, #53654)
+
     * ``bzr help commands`` output is now shorter (Aaron Bentley)
 
+    * New connection: ``bzr+http://`` which supports tunnelling the smart
+      protocol over an HTTP connection. If writing is enabled on the bzr
+      server, then you can write over the http connection.
+      (Andrew Bennetts)
+
     * ``bzr`` now uses lazy importing to reduce the startup time. This has
       a moderate effect on lots of actions, especially ones that have
       little to do. For example ``bzr rocks`` time is down to 116ms from
       283ms. (John Arbash Meinel)
 
+    * New Registry class to provide name-to-object registry-like support,
+      for example for schemes where plugins can register new classes to
+      do certain tasks (e.g. log formatters). Also provides lazy registration
+      to allow modules to be loaded on request. (John Arbash Meinel, Adeodato
+      Simó)
+
   INTERNALS:
 
     * ``TestCase.run_bzr``, ``run_bzr_captured``, and ``run_bzr_subprocess``
@@ -18,6 +33,10 @@
       around a regex, which defers compilation until first use. 
       (John Arbash Meinel)
 
+    * ``TestCase.run_bzr_subprocess`` defaults to supplying the
+      ``--no-plugins`` parameter to ensure test reproducability, and avoid
+      problems with system-wide installed plugins. (John Arbash Meinel)
+
   BUG FIXES:
 
     * Give nicer error message when an http server returns a 403
@@ -57,6 +76,10 @@
     * TestingHTTPRequestHandler really handles the Range header
       (previously it was ignoring it and returning the whole file,).
 
+    * ``TestCase.make_branch_and_memory_tree`` now takes a format
+      option to set the BzrDir, Repository and Branch formats of the
+      created objects. (Robert Collins, John Arbash Meinel)
+
 bzr 0.11  2006-10-02
 
     * Smart server transport test failures on windows fixed. (Lukáš Lalinský).
@@ -74,9 +97,6 @@
 
   IMPROVEMENTS:
 
-    * urllib implements keep-alive connections and share them
-      (Vincent Ladeuil, #53654)
-
     * Knit files now wait to create their contents until the first data is
       added. The old code used to create an empty .knit and a .kndx with just
       the header. However, this caused a lot of extra round trips over sftp.

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2006-10-11 00:23:23 +0000
+++ bzrlib/builtins.py	2006-10-13 06:41:04 +0000
@@ -2855,7 +2855,7 @@
             url = 'readonly+' + url
         t = get_transport(url)
         if inet:
-            server = smart.SmartStreamServer(sys.stdin, sys.stdout, t)
+            server = smart.SmartServerPipeStreamMedium(sys.stdin, sys.stdout, t)
         elif port is not None:
             if ':' in port:
                 host, port = port.split(':')

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-10-05 09:31:55 +0000
+++ bzrlib/errors.py	2006-10-13 10:50:33 +0000
@@ -270,6 +270,16 @@
     """Directory not empty: %(path)r%(extra)s"""
 
 
+class ReadingCompleted(BzrNewError):
+    """The MediumRequest '%(request)s' has already had finish_reading called upon it - the request has been completed and no more data may be read."""
+
+    is_user_error = False
+
+    def __init__(self, request):
+        BzrNewError.__init__(self)
+        self.request = request
+
+
 class ResourceBusy(PathError):
     """Device or resource busy: %(path)r%(extra)s"""
 
@@ -802,11 +812,21 @@
 
 class NoSuchExportFormat(BzrNewError):
     """Export format %(format)r not supported"""
+
     def __init__(self, format):
         BzrNewError.__init__(self)
         self.format = format
 
 
+
+class TooManyConcurrentRequests(BzrNewError):
+    """The medium '%(medium)s' has reached its concurrent request limit. Be sure to finish_writing and finish_reading on the current request that is open."""
+
+    def __init__(self, medium):
+        BzrNewError.__init__(self)
+        self.medium = medium
+
+
 class TransportError(BzrNewError):
     """Transport error: %(msg)s %(orig_error)s"""
 
@@ -911,6 +931,26 @@
                           " unchanged." % tree.basedir)
 
 
+class WritingCompleted(BzrNewError):
+    """The MediumRequest '%(request)s' has already had finish_writing called upon it - accept bytes may not be called anymore."""
+
+    is_user_error = False
+
+    def __init__(self, request):
+        BzrNewError.__init__(self)
+        self.request = request
+
+
+class WritingNotComplete(BzrNewError):
+    """The MediumRequest '%(request)s' has not has finish_writing called upon it - until the write phase is complete no data may be read."""
+
+    is_user_error = False
+
+    def __init__(self, request):
+        BzrNewError.__init__(self)
+        self.request = request
+
+
 class CantReprocessAndShowBase(BzrNewError):
     """Can't reprocess and show base.
 Reprocessing obscures relationship of conflicting lines to base."""
@@ -931,6 +971,14 @@
         self.filename = filename
 
 
+class MediumNotConnected(BzrNewError):
+    """The medium '%(medium)s' is not connected."""
+
+    def __init__(self, medium):
+        BzrNewError.__init__(self)
+        self.medium = medium
+
+
 class MustUseDecorated(Exception):
     """A decorating function has requested its original command be used.
     
@@ -1224,6 +1272,14 @@
         BadInventoryFormat.__init__(self, msg=msg)
 
 
+class NoSmartMedium(BzrNewError):
+    """The transport '%(transport)s' cannot tunnel the smart protocol."""
+
+    def __init__(self, transport):
+        BzrNewError.__init__(self)
+        self.transport = transport
+
+
 class NoSmartServer(NotBranchError):
     """No smart server available at %(url)s"""
 

=== added file 'bzrlib/registry.py'
--- bzrlib/registry.py	1970-01-01 00:00:00 +0000
+++ bzrlib/registry.py	2006-10-13 00:55:26 +0000
@@ -0,0 +1,214 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Classes to provide name-to-object registry-like support."""
+
+
+class _ObjectGetter(object):
+    """Maintain a reference to an object, and return the object on request.
+
+    This is used by Registry to make plain objects function similarly
+    to lazily imported objects.
+
+    Objects can be any sort of python object (class, function, module,
+    instance, etc)
+    """
+
+    __slots__ = ['_obj']
+
+    def __init__(self, obj):
+        self._obj = obj
+
+    def get_obj(self):
+        """Get the object that was saved at creation time"""
+        return self._obj
+
+
+class _LazyObjectGetter(_ObjectGetter):
+    """Keep a record of a possible object.
+
+    When requested, load and return it.
+    """
+
+    __slots__ = ['_module_name', '_member_name', '_imported']
+
+    def __init__(self, module_name, member_name):
+        self._module_name = module_name
+        self._member_name = member_name
+        self._imported = False
+        super(_LazyObjectGetter, self).__init__(None)
+
+    def get_obj(self):
+        """Get the referenced object.
+
+        Upon first request, the object will be imported. Future requests will
+        return the imported object.
+        """
+        if not self._imported:
+            self._do_import()
+        return super(_LazyObjectGetter, self).get_obj()
+
+    def _do_import(self):
+        obj = __import__(self._module_name, globals(), locals(),
+                         [self._member_name])
+        if self._member_name:
+            obj = getattr(obj, self._member_name)
+        self._obj = obj
+        self._imported = True
+
+
+class Registry(object):
+    """A class that registers objects to a name.
+
+    There are many places that want to collect related objects and access them
+    by a key. This class is designed to allow registering the mapping from key
+    to object. It goes one step further, and allows registering a name to a
+    hypothetical object which has not been imported yet. It also supports
+    adding additional information at registration time so that decisions can be
+    made without having to import the object (which may be expensive).
+
+    The functions 'get', 'get_info', and 'get_help' also support a
+    'default_key' (settable through my_registry.default_key = XXX, XXX must
+    already be registered.) Calling my_registry.get() or my_registry.get(None),
+    will return the entry for the default key.
+    """
+
+    def __init__(self):
+        """Create a new Registry."""
+        self._default_key = None
+        # Map from key => (is_lazy, info)
+        self._dict = {}
+        self._help_dict = {}
+        self._info_dict = {}
+
+    def register(self, key, obj, help=None, info=None,
+                 override_existing=False):
+        """Register a new object to a name.
+
+        :param key: This is the key to use to request the object later.
+        :param obj: The object to register.
+        :param help: Help text for this entry. This may be a string or
+                a callable. If it is a callable, it should take two
+                parameters (registry, key): this registry and the key that 
+                the help was registered under.
+        :param info: More information for this entry. Registry.get_info()
+                can be used to get this information. Registry treats this as an
+                opaque storage location (it is defined by the caller).
+        :param override_existing: Raise KeyErorr if False and something has
+                already been registered for that key. If True, ignore if there
+                is an existing key (always register the new value).
+        """
+        if not override_existing:
+            if key in self._dict:
+                raise KeyError('Key %r already registered' % key)
+        self._dict[key] = _ObjectGetter(obj)
+        self._add_help_and_info(key, help=help, info=info)
+
+    def register_lazy(self, key, module_name, member_name,
+                      help=None, info=None,
+                      override_existing=False):
+        """Register a new object to be loaded on request.
+
+        :param module_name: The python path to the module. Such as 'os.path'.
+        :param member_name: The member of the module to return.  If empty or 
+                None, get() will return the module itself.
+        :param help: Help text for this entry. This may be a string or
+                a callable.
+        :param info: More information for this entry. Registry 
+        :param override_existing: If True, replace the existing object
+                with the new one. If False, if there is already something
+                registered with the same key, raise a KeyError
+        """
+        if not override_existing:
+            if key in self._dict:
+                raise KeyError('Key %r already registered' % key)
+        self._dict[key] = _LazyObjectGetter(module_name, member_name)
+        self._add_help_and_info(key, help=help, info=info)
+
+    def _add_help_and_info(self, key, help=None, info=None):
+        """Add the help and information about this key"""
+        self._help_dict[key] = help
+        self._info_dict[key] = info
+
+    def get(self, key=None):
+        """Return the object register()'ed to the given key.
+
+        May raise ImportError if the object was registered lazily and
+        there are any problems, or AttributeError if the module does not 
+        have the supplied member.
+
+        :param key: The key to obtain the object for. If no object has been
+            registered to that key, the object registered for self.default_key
+            will be returned instead, if it exists. Otherwise KeyError will be
+            raised.
+        :return: The previously registered object.
+        :raises ImportError: If the object was registered lazily, and there are
+            problems during import.
+        :raises AttributeError: If registered lazily, and the module does not
+            contain the registered member.
+        """
+        return self._dict[self._get_key_or_default(key)].get_obj()
+
+    def _get_key_or_default(self, key=None):
+        """Return either 'key' or the default key if key is None"""
+        if key is not None:
+            return key
+        if self.default_key is None:
+            raise KeyError('Key is None, and no default key is set')
+        else:
+            return self.default_key
+
+    def get_help(self, key=None):
+        """Get the help text associated with the given key"""
+        the_help = self._help_dict[self._get_key_or_default(key)]
+        if callable(the_help):
+            return the_help(self, key)
+        return the_help
+
+    def get_info(self, key=None):
+        """Get the extra information associated with the given key"""
+        return self._info_dict[self._get_key_or_default(key)]
+
+    def remove(self, key):
+        """Remove a registered entry.
+
+        This is mostly for the test suite, but it can be used by others
+        """
+        del self._dict[key]
+
+    def __contains__(self, key):
+        return key in self._dict
+
+    def keys(self):
+        """Get a list of registered entries"""
+        return sorted(self._dict.keys())
+
+    def iteritems(self):
+        for key, getter in self._dict.iteritems():
+            yield key, getter.get_obj()
+
+    def _set_default_key(self, key):
+        if not self._dict.has_key(key):
+            raise KeyError('No object registered under key %s.' % key)
+        else:
+            self._default_key = key
+
+    def _get_default_key(self):
+        return self._default_key
+
+    default_key = property(_get_default_key, _set_default_key,
+                            doc="Current value of the default key."
+                                " Can be set to any existing key.")

=== modified file 'bzrlib/tests/HTTPTestUtil.py'
--- bzrlib/tests/HTTPTestUtil.py	2006-10-13 08:53:34 +0000
+++ bzrlib/tests/HTTPTestUtil.py	2006-10-13 10:50:33 +0000
@@ -14,6 +14,7 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
+from cStringIO import StringIO
 import errno
 import socket
 
@@ -22,6 +23,10 @@
     HttpServer,
     TestingHTTPRequestHandler,
     )
+from bzrlib.transport import (
+    get_transport,
+    smart,
+    )
 
 
 class WallRequestHandler(TestingHTTPRequestHandler):
@@ -93,6 +98,48 @@
         return False
 
 
+class HTTPServerWithSmarts(HttpServer):
+    """HTTPServerWithSmarts extends the HttpServer with POST methods that will
+    trigger a smart server to execute with a transport rooted at the rootdir of
+    the HTTP server.
+    """
+
+    def __init__(self):
+        HttpServer.__init__(self, SmartRequestHandler)
+
+
+class SmartRequestHandler(TestingHTTPRequestHandler):
+    """Extend TestingHTTPRequestHandler to support smart client POSTs."""
+
+    def do_POST(self):
+        """Hand the request off to a smart server instance."""
+        self.send_response(200)
+        self.send_header("Content-type", "application/octet-stream")
+        transport = get_transport(self.server.test_case._home_dir)
+        # TODO: We might like to support streaming responses.  1.0 allows no
+        # Content-length in this case, so for integrity we should perform our
+        # own chunking within the stream.
+        # 1.1 allows chunked responses, and in this case we could chunk using
+        # the HTTP chunking as this will allow HTTP persistence safely, even if
+        # we have to stop early due to error, but we would also have to use the
+        # HTTP trailer facility which may not be widely available.
+        out_buffer = StringIO()
+        smart_protocol_request = smart.SmartServerRequestProtocolOne(
+                transport, out_buffer.write)
+        # if this fails, we should return 400 bad request, but failure is
+        # failure for now - RBC 20060919
+        data_length = int(self.headers['Content-Length'])
+        # Perhaps there should be a SmartServerHTTPMedium that takes care of
+        # feeding the bytes in the http request to the smart_protocol_request,
+        # but for now it's simpler to just feed the bytes directly.
+        smart_protocol_request.accept_bytes(self.rfile.read(data_length))
+        assert smart_protocol_request.next_read_size() == 0, (
+            "not finished reading, but all data sent to protocol.")
+        self.send_header("Content-Length", str(len(out_buffer.getvalue())))
+        self.end_headers()
+        self.wfile.write(out_buffer.getvalue())
+
+
 class TestCaseWithWebserver(TestCaseWithTransport):
     """A support class that provides readonly urls that are http://.
 

=== modified file 'bzrlib/tests/HttpServer.py'
--- bzrlib/tests/HttpServer.py	2006-10-12 14:29:32 +0000
+++ bzrlib/tests/HttpServer.py	2006-10-13 10:50:33 +0000
@@ -237,11 +237,14 @@
         Server.__init__(self)
         self.request_handler = request_handler
 
+    def _get_httpd(self):
+        return TestingHTTPServer(('localhost', 0),
+                                  self.request_handler,
+                                  self)
+
     def _http_start(self):
         httpd = None
-        httpd = TestingHTTPServer(('localhost', 0),
-                                  self.request_handler,
-                                  self)
+        httpd = self._get_httpd()
         host, port = httpd.socket.getsockname()
         self._http_base_url = '%s://localhost:%s/' % (self._url_protocol, port)
         self._http_starting.release()

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2006-10-12 14:29:32 +0000
+++ bzrlib/tests/__init__.py	2006-10-13 10:50:33 +0000
@@ -928,11 +928,17 @@
             The values must be strings. The change will only occur in the
             child, so you don't need to fix the environment after running.
         :param universal_newlines: Convert CRLF => LF
+        :param allow_plugins: By default the subprocess is run with
+            --no-plugins to ensure test reproducibility. Also, it is possible
+            for system-wide plugins to create unexpected output on stderr,
+            which can cause unnecessary test failures.
         """
         env_changes = kwargs.get('env_changes', {})
         working_dir = kwargs.get('working_dir', None)
+        allow_plugins = kwargs.get('allow_plugins', False)
         process = self.start_bzr_subprocess(args, env_changes=env_changes,
-                                            working_dir=working_dir)
+                                            working_dir=working_dir,
+                                            allow_plugins=allow_plugins)
         # We distinguish between retcode=None and retcode not passed.
         supplied_retcode = kwargs.get('retcode', 0)
         return self.finish_bzr_subprocess(process, retcode=supplied_retcode,
@@ -941,7 +947,8 @@
 
     def start_bzr_subprocess(self, process_args, env_changes=None,
                              skip_if_plan_to_signal=False,
-                             working_dir=None):
+                             working_dir=None,
+                             allow_plugins=False):
         """Start bzr in a subprocess for testing.
 
         This starts a new Python interpreter and runs bzr in there.
@@ -958,6 +965,7 @@
             child, so you don't need to fix the environment after running.
         :param skip_if_plan_to_signal: raise TestSkipped when true and os.kill
             is not available.
+        :param allow_plugins: If False (default) pass --no-plugins to bzr.
 
         :returns: Popen object for the started process.
         """
@@ -989,8 +997,11 @@
             # so we will avoid using it on all platforms, just to
             # make sure the code path is used, and we don't break on win32
             cleanup_environment()
-            process = Popen([sys.executable, bzr_path] + list(process_args),
-                             stdin=PIPE, stdout=PIPE, stderr=PIPE)
+            command = [sys.executable, bzr_path]
+            if not allow_plugins:
+                command.append('--no-plugins')
+            command.extend(process_args)
+            process = self._popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE)
         finally:
             restore_environment()
             if cwd is not None:
@@ -998,6 +1009,14 @@
 
         return process
 
+    def _popen(self, *args, **kwargs):
+        """Place a call to Popen.
+
+        Allows tests to override this method to intercept the calls made to
+        Popen for introspection.
+        """
+        return Popen(*args, **kwargs)
+
     def get_bzr_path(self):
         """Return the path of the 'bzr' executable for this test suite."""
         bzr_path = os.path.dirname(os.path.dirname(bzrlib.__file__))+'/bzr'
@@ -1297,9 +1316,9 @@
         made_control = self.make_bzrdir(relpath, format=format)
         return made_control.create_repository(shared=shared)
 
-    def make_branch_and_memory_tree(self, relpath):
+    def make_branch_and_memory_tree(self, relpath, format=None):
         """Create a branch on the default transport and a MemoryTree for it."""
-        b = self.make_branch(relpath)
+        b = self.make_branch(relpath, format=format)
         return memorytree.MemoryTree.create_on_branch(b)
 
     def overrideEnvironmentForTesting(self):
@@ -1648,6 +1667,7 @@
                    'bzrlib.tests.test_plugins',
                    'bzrlib.tests.test_progress',
                    'bzrlib.tests.test_reconcile',
+                   'bzrlib.tests.test_registry',
                    'bzrlib.tests.test_repository',
                    'bzrlib.tests.test_revert',
                    'bzrlib.tests.test_revision',

=== modified file 'bzrlib/tests/blackbox/test_selftest.py'
--- bzrlib/tests/blackbox/test_selftest.py	2006-09-28 21:53:09 +0000
+++ bzrlib/tests/blackbox/test_selftest.py	2006-10-13 07:41:06 +0000
@@ -359,6 +359,32 @@
         self.assertEqual(cwd, osutils.getcwd())
 
 
+class _DontSpawnProcess(Exception):
+    """A simple exception which just allows us to skip unnecessary steps"""
+
+
+class TestRunBzrSubprocessCommands(TestCaseWithTransport):
+
+    def _popen(self, *args, **kwargs):
+        """Record the command that is run, so that we can ensure it is correct"""
+        self._popen_args = args
+        self._popen_kwargs = kwargs
+        raise _DontSpawnProcess()
+
+    def test_run_bzr_subprocess_no_plugins(self):
+        self.assertRaises(_DontSpawnProcess, self.run_bzr_subprocess)
+        command = self._popen_args[0]
+        self.assertEqual(sys.executable, command[0])
+        self.assertEqual(self.get_bzr_path(), command[1])
+        self.assertEqual(['--no-plugins'], command[2:])
+
+    def test_allow_plugins(self):
+        self.assertRaises(_DontSpawnProcess,
+                          self.run_bzr_subprocess, allow_plugins=True)
+        command = self._popen_args[0]
+        self.assertEqual([], command[2:])
+
+
 class TestBzrSubprocess(TestCaseWithTransport):
 
     def test_start_and_stop_bzr_subprocess(self):

=== modified file 'bzrlib/tests/blackbox/test_serve.py'
--- bzrlib/tests/blackbox/test_serve.py	2006-09-19 03:09:24 +0000
+++ bzrlib/tests/blackbox/test_serve.py	2006-09-21 09:15:20 +0000
@@ -30,31 +30,13 @@
 from bzrlib.transport import get_transport, smart
 
 
-class DoesNotCloseStdOutClient(smart.SmartStreamClient):
-    """A client that doesn't close stdout upon disconnect().
-    
-    We wish to let stdout remain open so that we can see if the server writes
-    anything to stdout during its shutdown.
-    """
-
-    def disconnect(self):
-        if self._connected:
-            self._connected = False
-            # The client's out is the server's in.
-            self._out.close()
-
-
 class TestBzrServe(TestCaseWithTransport):
 
-    def assertInetServerShutsdownCleanly(self, client, process):
+    def assertInetServerShutsdownCleanly(self, process):
         """Shutdown the server process looking for errors."""
-        # Disconnect the client forcefully JUST IN CASE because of __del__'s use
-        # in the smart module.
-        client.disconnect()
-
-        # Shutdown the server: the client should have disconnected cleanly and
-        # closed stdin, so the server process should shut itself down.
-        self.assertTrue(process.stdin.closed)
+        # Shutdown the server: the server should shut down when it cannot read
+        # from stdin anymore.
+        process.stdin.close()
         # Hide stdin from the subprocess module, so it won't fail to close it.
         process.stdin = None
         result = self.finish_bzr_subprocess(process, retcode=0)
@@ -82,10 +64,10 @@
         # Connect to the server
         # We use this url because while this is no valid URL to connect to this
         # server instance, the transport needs a URL.
-        client = DoesNotCloseStdOutClient(
-            lambda: (process.stdout, process.stdin))
-        transport = smart.SmartTransport('bzr://localhost/', client=client)
-        return process, client, transport
+        medium = smart.SmartSimplePipesClientMedium(
+            process.stdout, process.stdin)
+        transport = smart.SmartTransport('bzr://localhost/', medium=medium)
+        return process, transport
 
     def start_server_port(self, extra_options=()):
         """Start a bzr server subprocess.
@@ -106,27 +88,21 @@
 
     def test_bzr_serve_inet_readonly(self):
         """bzr server should provide a read only filesystem by default."""
-        process, client, transport = self.start_server_inet()
+        process, transport = self.start_server_inet()
         self.assertRaises(errors.TransportNotPossible, transport.mkdir, 'adir')
-        # finish with the transport
-        del transport
-        self.assertInetServerShutsdownCleanly(client, process)
+        self.assertInetServerShutsdownCleanly(process)
 
     def test_bzr_serve_inet_readwrite(self):
         # Make a branch
         self.make_branch('.')
 
-        process, client, transport = self.start_server_inet(['--allow-writes'])
+        process, transport = self.start_server_inet(['--allow-writes'])
 
         # We get a working branch
         branch = BzrDir.open_from_transport(transport).open_branch()
         branch.repository.get_revision_graph()
         self.assertEqual(None, branch.last_revision())
-
-        # finish with the transport
-        del transport
-
-        self.assertInetServerShutsdownCleanly(client, process)
+        self.assertInetServerShutsdownCleanly(process)
 
     def test_bzr_serve_port_readonly(self):
         """bzr server should provide a read only filesystem by default."""

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2006-09-27 19:20:35 +0000
+++ bzrlib/tests/test_errors.py	2006-09-28 04:01:35 +0000
@@ -26,18 +26,36 @@
 
 class TestErrors(TestCaseWithTransport):
 
+    def test_medium_not_connected(self):
+        error = errors.MediumNotConnected("a medium")
+        self.assertEqualDiff(
+            "The medium 'a medium' is not connected.", str(error))
+        
     def test_no_repo(self):
         dir = bzrdir.BzrDir.create(self.get_url())
         error = errors.NoRepositoryPresent(dir)
         self.assertNotEqual(-1, str(error).find((dir.transport.clone('..').base)))
         self.assertEqual(-1, str(error).find((dir.transport.base)))
         
+    def test_no_smart_medium(self):
+        error = errors.NoSmartMedium("a transport")
+        self.assertEqualDiff("The transport 'a transport' cannot tunnel the "
+            "smart protocol.",
+            str(error))
+
     def test_no_such_id(self):
         error = errors.NoSuchId("atree", "anid")
         self.assertEqualDiff("The file id anid is not present in the tree "
             "atree.",
             str(error))
 
+    def test_too_many_concurrent_requests(self):
+        error = errors.TooManyConcurrentRequests("a medium")
+        self.assertEqualDiff("The medium 'a medium' has reached its concurrent "
+            "request limit. Be sure to finish_writing and finish_reading on "
+            "the current request that is open.",
+            str(error))
+
     def test_up_to_date(self):
         error = errors.UpToDateFormat(bzrdir.BzrDirFormat4())
         self.assertEqualDiff("The branch format Bazaar-NG branch, "
@@ -53,6 +71,27 @@
                              repo.bzrdir.root_transport.base,
                              str(error))
 
+    def test_reading_completed(self):
+        error = errors.ReadingCompleted("a request")
+        self.assertEqualDiff("The MediumRequest 'a request' has already had "
+            "finish_reading called upon it - the request has been completed and"
+            " no more data may be read.",
+            str(error))
+
+    def test_writing_completed(self):
+        error = errors.WritingCompleted("a request")
+        self.assertEqualDiff("The MediumRequest 'a request' has already had "
+            "finish_writing called upon it - accept bytes may not be called "
+            "anymore.",
+            str(error))
+
+    def test_writing_not_completed(self):
+        error = errors.WritingNotComplete("a request")
+        self.assertEqualDiff("The MediumRequest 'a request' has not has "
+            "finish_writing called upon it - until the write phase is complete"
+            " no data may be read.",
+            str(error))
+
 
 class PassThroughError(errors.BzrNewError):
     """Pass through %(foo)s and %(bar)s"""

=== modified file 'bzrlib/tests/test_http.py'
--- bzrlib/tests/test_http.py	2006-10-13 08:53:34 +0000
+++ bzrlib/tests/test_http.py	2006-10-13 10:50:33 +0000
@@ -19,7 +19,9 @@
 
 # TODO: Should be renamed to bzrlib.transport.http.tests?
 
+import select
 import socket
+import threading
 
 import bzrlib
 from bzrlib import errors
@@ -51,7 +53,7 @@
 from bzrlib.transport.http._urllib import HttpTransport_urllib
 
 
-class FakeManager (object):
+class FakeManager(object):
 
     def __init__(self):
         self.credentials = []
@@ -60,6 +62,60 @@
         self.credentials.append([realm, host, username, password])
 
 
+class RecordingServer(object):
+    """A fake HTTP server.
+    
+    It records the bytes sent to it, and replies with a 200.
+    """
+
+    def __init__(self, expect_body_tail=None):
+        """Constructor.
+
+        :type expect_body_tail: str
+        :param expect_body_tail: a reply won't be sent until this string is
+            received.
+        """
+        self._expect_body_tail = expect_body_tail
+        self.host = None
+        self.port = None
+        self.received_bytes = ''
+
+    def setUp(self):
+        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        self._sock.bind(('127.0.0.1', 0))
+        self.host, self.port = self._sock.getsockname()
+        self._ready = threading.Event()
+        self._thread = threading.Thread(target=self._accept_read_and_reply)
+        self._thread.setDaemon(True)
+        self._thread.start()
+        self._ready.wait(5)
+
+    def _accept_read_and_reply(self):
+        self._sock.listen(1)
+        self._ready.set()
+        self._sock.settimeout(5)
+        try:
+            conn, address = self._sock.accept()
+            # On win32, the accepted connection will be non-blocking to start
+            # with because we're using settimeout.
+            conn.setblocking(True)
+            while not self.received_bytes.endswith(self._expect_body_tail):
+                self.received_bytes += conn.recv(4096)
+            conn.sendall('HTTP/1.1 200 OK\r\n')
+        except socket.timeout:
+            # Make sure the client isn't stuck waiting for us to e.g. accept.
+            self._sock.close()
+
+    def tearDown(self):
+        try:
+            self._sock.close()
+        except socket.error:
+            # We might have already closed it.  We don't care.
+            pass
+        self.host = None
+        self.port = None
+
+
 class TestHttpUrls(TestCase):
 
     def test_url_parsing(self):
@@ -154,6 +210,13 @@
             '"GET /foo/bar HTTP/1.1" 200 - "-" "bzr/%s'
             % bzrlib.__version__) > -1)
 
+    def test_get_smart_medium(self):
+        # For HTTP, get_smart_medium should return the transport object.
+        server = self.get_readonly_server()
+        http_transport = self._transport(server.get_url())
+        medium = http_transport.get_smart_medium()
+        self.assertIs(medium, http_transport)
+
     def test_has_on_bogus_host(self):
         # Get a free address and don't 'accept' on it, so that we
         # can be sure there is no http handler there, but set a
@@ -228,6 +291,34 @@
         self.assertEqual([[10, 12], [22, 26]], ranges)
 
 
+class TestPost(TestCase):
+
+    def _test_post_body_is_received(self, scheme):
+        server = RecordingServer(expect_body_tail='end-of-body')
+        server.setUp()
+        self.addCleanup(server.tearDown)
+        url = '%s://%s:%s/' % (scheme, server.host, server.port)
+        try:
+            http_transport = get_transport(url)
+        except UnsupportedProtocol:
+            raise TestSkipped('%s not available' % scheme)
+        code, response = http_transport._post('abc def end-of-body')
+        self.assertTrue(
+            server.received_bytes.startswith('POST /.bzr/smart HTTP/1.'))
+        self.assertTrue('content-length: 19\r' in server.received_bytes.lower())
+        # The transport should not be assuming that the server can accept
+        # chunked encoding the first time it connects, because HTTP/1.1, so we
+        # check for the literal string.
+        self.assertTrue(
+            server.received_bytes.endswith('\r\n\r\nabc def end-of-body'))
+
+    def test_post_body_is_received_urllib(self):
+        self._test_post_body_is_received('http+urllib')
+
+    def test_post_body_is_received_pycurl(self):
+        self._test_post_body_is_received('http+pycurl')
+
+
 class TestRangeHeader(TestCase):
     """Test range_header method"""
 
@@ -392,3 +483,33 @@
                                  TestCaseWithWebserver):
     """Tests forbidden server for pycurl implementation"""
 
+
+class TestRecordingServer(TestCase):
+
+    def test_create(self):
+        server = RecordingServer(expect_body_tail=None)
+        self.assertEqual('', server.received_bytes)
+        self.assertEqual(None, server.host)
+        self.assertEqual(None, server.port)
+
+    def test_setUp_and_tearDown(self):
+        server = RecordingServer(expect_body_tail=None)
+        server.setUp()
+        try:
+            self.assertNotEqual(None, server.host)
+            self.assertNotEqual(None, server.port)
+        finally:
+            server.tearDown()
+        self.assertEqual(None, server.host)
+        self.assertEqual(None, server.port)
+
+    def test_send_receive_bytes(self):
+        server = RecordingServer(expect_body_tail='c')
+        server.setUp()
+        self.addCleanup(server.tearDown)
+        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        sock.connect((server.host, server.port))
+        sock.sendall('abc')
+        self.assertEqual('HTTP/1.1 200 OK\r\n',
+                         sock.recv(4096, socket.MSG_WAITALL))
+        self.assertEqual('abc', server.received_bytes)

=== added file 'bzrlib/tests/test_registry.py'
--- bzrlib/tests/test_registry.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_registry.py	2006-09-21 22:25:19 +0000
@@ -0,0 +1,262 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for the Registry classes"""
+
+import os
+import sys
+
+from bzrlib import (
+    errors,
+    registry,
+    osutils,
+    )
+from bzrlib.tests import TestCase, TestCaseInTempDir
+
+
+class TestRegistry(TestCase):
+
+    def register_stuff(self, a_registry):
+        a_registry.register('one', 1)
+        a_registry.register('two', 2)
+        a_registry.register('four', 4)
+        a_registry.register('five', 5)
+
+    def test_registry(self):
+        a_registry = registry.Registry()
+        self.register_stuff(a_registry)
+
+        self.failUnless(a_registry.default_key is None)
+
+        # test get() (self.default_key == None)
+        self.assertRaises(KeyError, a_registry.get)
+        self.assertRaises(KeyError, a_registry.get, None)
+        self.assertEqual(2, a_registry.get('two'))
+        self.assertRaises(KeyError, a_registry.get, 'three')
+
+        # test _set_default_key
+        a_registry.default_key = 'five'
+        self.failUnless(a_registry.default_key == 'five')
+        self.assertEqual(5, a_registry.get())
+        self.assertEqual(5, a_registry.get(None))
+        # If they ask for a specific entry, they should get KeyError
+        # not the default value. They can always pass None if they prefer
+        self.assertRaises(KeyError, a_registry.get, 'six')
+        self.assertRaises(KeyError, a_registry._set_default_key, 'six')
+
+        # test keys()
+        self.assertEqual(['five', 'four', 'one', 'two'], a_registry.keys())
+
+    def test_registry_funcs(self):
+        a_registry = registry.Registry()
+        self.register_stuff(a_registry)
+
+        self.failUnless('one' in a_registry)
+        a_registry.remove('one')
+        self.failIf('one' in a_registry)
+        self.assertRaises(KeyError, a_registry.get, 'one')
+
+        a_registry.register('one', 'one')
+
+        self.assertEqual(['five', 'four', 'one', 'two'],
+                         sorted(a_registry.keys()))
+        self.assertEqual([('five', 5), ('four', 4),
+                          ('one', 'one'), ('two', 2)],
+                         sorted(a_registry.iteritems()))
+
+    def test_register_override(self):
+        a_registry = registry.Registry()
+        a_registry.register('one', 'one')
+        self.assertRaises(KeyError, a_registry.register, 'one', 'two')
+        self.assertRaises(KeyError, a_registry.register, 'one', 'two',
+                                    override_existing=False)
+
+        a_registry.register('one', 'two', override_existing=True)
+        self.assertEqual('two', a_registry.get('one'))
+
+        self.assertRaises(KeyError, a_registry.register_lazy,
+                          'one', 'three', 'four')
+
+        a_registry.register_lazy('one', 'module', 'member',
+                                 override_existing=True)
+
+    def test_registry_help(self):
+        a_registry = registry.Registry()
+        a_registry.register('one', 1, help='help text for one')
+        # We should not have to import the module to return the help
+        # information
+        a_registry.register_lazy('two', 'nonexistent_module', 'member',
+                                 help='help text for two')
+
+        # We should be able to handle a callable to get information
+        help_calls = []
+        def generic_help(reg, key):
+            help_calls.append(key)
+            return 'generic help for %s' % (key,)
+        a_registry.register('three', 3, help=generic_help)
+        a_registry.register_lazy('four', 'nonexistent_module', 'member2',
+                                 help=generic_help)
+        a_registry.register('five', 5)
+
+        def help_from_object(reg, key):
+            obj = reg.get(key)
+            return obj.help()
+
+        class SimpleObj(object):
+            def help(self):
+                return 'this is my help'
+        a_registry.register('six', SimpleObj(), help=help_from_object)
+
+        self.assertEqual('help text for one', a_registry.get_help('one'))
+        self.assertEqual('help text for two', a_registry.get_help('two'))
+        self.assertEqual('generic help for three',
+                         a_registry.get_help('three'))
+        self.assertEqual(['three'], help_calls)
+        self.assertEqual('generic help for four',
+                         a_registry.get_help('four'))
+        self.assertEqual(['three', 'four'], help_calls)
+        self.assertEqual(None, a_registry.get_help('five'))
+        self.assertEqual('this is my help', a_registry.get_help('six'))
+
+        self.assertRaises(KeyError, a_registry.get_help, None)
+        self.assertRaises(KeyError, a_registry.get_help, 'seven')
+
+        a_registry.default_key = 'one'
+        self.assertEqual('help text for one', a_registry.get_help(None))
+        self.assertRaises(KeyError, a_registry.get_help, 'seven')
+
+        self.assertEqual([('five', None),
+                          ('four', 'generic help for four'),
+                          ('one', 'help text for one'),
+                          ('six', 'this is my help'),
+                          ('three', 'generic help for three'),
+                          ('two', 'help text for two'),
+                         ], sorted((key, a_registry.get_help(key))
+                                    for key in a_registry.keys()))
+
+        # We don't know what order it was called in, but we should get
+        # 2 more calls to three and four
+        self.assertEqual(['four', 'four', 'three', 'three'],
+                         sorted(help_calls))
+
+    def test_registry_info(self):
+        a_registry = registry.Registry()
+        a_registry.register('one', 1, info='string info')
+        # We should not have to import the module to return the info
+        a_registry.register_lazy('two', 'nonexistent_module', 'member',
+                                 info=2)
+
+        # We should be able to handle a callable to get information
+        a_registry.register('three', 3, info=['a', 'list'])
+        obj = object()
+        a_registry.register_lazy('four', 'nonexistent_module', 'member2',
+                                 info=obj)
+        a_registry.register('five', 5)
+
+        self.assertEqual('string info', a_registry.get_info('one'))
+        self.assertEqual(2, a_registry.get_info('two'))
+        self.assertEqual(['a', 'list'], a_registry.get_info('three'))
+        self.assertIs(obj, a_registry.get_info('four'))
+        self.assertIs(None, a_registry.get_info('five'))
+
+        self.assertRaises(KeyError, a_registry.get_info, None)
+        self.assertRaises(KeyError, a_registry.get_info, 'six')
+
+        a_registry.default_key = 'one'
+        self.assertEqual('string info', a_registry.get_info(None))
+        self.assertRaises(KeyError, a_registry.get_info, 'six')
+
+        self.assertEqual([('five', None),
+                          ('four', obj),
+                          ('one', 'string info'),
+                          ('three', ['a', 'list']),
+                          ('two', 2),
+                         ], sorted((key, a_registry.get_info(key))
+                                    for key in a_registry.keys()))
+
+
+class TestRegistryWithDirs(TestCaseInTempDir):
+    """Registry tests that require temporary dirs"""
+
+    def create_plugin_file(self, contents):
+        """Create a file to be used as a plugin.
+
+        This is created in a temporary directory, so that we
+        are sure that it doesn't start in the plugin path.
+        """
+        os.mkdir('tmp')
+        plugin_name = 'bzr_plugin_a_%s' % (osutils.rand_chars(4),)
+        open('tmp/'+plugin_name+'.py', 'wb').write(contents)
+        return plugin_name
+
+    def create_simple_plugin(self):
+        return self.create_plugin_file(
+            'object1 = "foo"\n'
+            '\n\n'
+            'def function(a,b,c):\n'
+            '    return a,b,c\n'
+            '\n\n'
+            'class MyClass(object):\n'
+            '    def __init__(self, a):\n'
+            '      self.a = a\n'
+            '\n\n'
+        )
+
+    def test_lazy_import_registry(self):
+        plugin_name = self.create_simple_plugin()
+        a_registry = registry.Registry()
+        a_registry.register_lazy('obj', plugin_name, 'object1')
+        a_registry.register_lazy('function', plugin_name, 'function')
+        a_registry.register_lazy('klass', plugin_name, 'MyClass')
+        a_registry.register_lazy('module', plugin_name, None)
+
+        self.assertEqual(['function', 'klass', 'module', 'obj'],
+                         sorted(a_registry.keys()))
+        # The plugin should not be loaded until we grab the first object
+        self.failIf(plugin_name in sys.modules)
+
+        # By default the plugin won't be in the search path
+        self.assertRaises(ImportError, a_registry.get, 'obj')
+
+        plugin_path = os.getcwd() + '/tmp'
+        sys.path.append(plugin_path)
+        try:
+            obj = a_registry.get('obj')
+            self.assertEqual('foo', obj)
+            self.failUnless(plugin_name in sys.modules)
+
+            # Now grab another object
+            func = a_registry.get('function')
+            self.assertEqual(plugin_name, func.__module__)
+            self.assertEqual('function', func.__name__)
+            self.assertEqual((1, [], '3'), func(1, [], '3'))
+
+            # And finally a class
+            klass = a_registry.get('klass')
+            self.assertEqual(plugin_name, klass.__module__)
+            self.assertEqual('MyClass', klass.__name__)
+
+            inst = klass(1)
+            self.assertIsInstance(inst, klass)
+            self.assertEqual(1, inst.a)
+
+            module = a_registry.get('module')
+            self.assertIs(obj, module.object1)
+            self.assertIs(func, module.function)
+            self.assertIs(klass, module.MyClass)
+        finally:
+            sys.path.remove(plugin_path)
+

=== modified file 'bzrlib/tests/test_selftest.py'
--- bzrlib/tests/test_selftest.py	2006-10-12 14:29:32 +0000
+++ bzrlib/tests/test_selftest.py	2006-10-13 10:50:33 +0000
@@ -23,8 +23,13 @@
 import unittest
 import warnings
 
-from bzrlib import osutils, memorytree
 import bzrlib
+from bzrlib import (
+    bzrdir,
+    memorytree,
+    osutils,
+    repository,
+    )
 from bzrlib.progress import _BaseProgressBar
 from bzrlib.tests import (
                           ChrootedTestCase,
@@ -465,6 +470,16 @@
         self.failIfExists('dir')
         self.assertIsInstance(tree, memorytree.MemoryTree)
 
+    def test_make_branch_and_memory_tree_with_format(self):
+        """make_branch_and_memory_tree should accept a format option."""
+        format = bzrdir.BzrDirMetaFormat1()
+        format.repository_format = repository.RepositoryFormat7()
+        tree = self.make_branch_and_memory_tree('dir', format=format)
+        self.failIfExists('dir')
+        self.assertIsInstance(tree, memorytree.MemoryTree)
+        self.assertEqual(format.repository_format.__class__,
+            tree.branch.repository._format.__class__)
+
 
 class TestTestCaseWithTransport(TestCaseWithTransport):
     """Tests for the convenience functions TestCaseWithTransport introduces."""

=== modified file 'bzrlib/tests/test_smart_transport.py'
--- bzrlib/tests/test_smart_transport.py	2006-09-28 10:31:13 +0000
+++ bzrlib/tests/test_smart_transport.py	2006-10-13 10:50:33 +0000
@@ -18,36 +18,504 @@
 
 # all of this deals with byte strings so this is safe
 from cStringIO import StringIO
-import subprocess
-import sys
+import os
+import socket
+import threading
+import urllib2
 
-import bzrlib
 from bzrlib import (
         bzrdir,
         errors,
+        osutils,
         tests,
         urlutils,
         )
+from bzrlib.tests.HTTPTestUtil import (
+        HTTPServerWithSmarts,
+        SmartRequestHandler,
+        )
 from bzrlib.transport import (
         get_transport,
         local,
         memory,
         smart,
         )
-
-
-class SmartClientTests(tests.TestCase):
-
-    def test_construct_smart_stream_client(self):
-        # make a new client; this really wants a connector function that returns
-        # two fifos or sockets but the constructor should not do any IO
-        client = smart.SmartStreamClient(None)
-
-
-class TCPClientTests(tests.TestCaseWithTransport):
+from bzrlib.transport.http import SmartClientHTTPMediumRequest
+
+
+class StringIOSSHVendor(object):
+    """A SSH vendor that uses StringIO to buffer writes and answer reads."""
+
+    def __init__(self, read_from, write_to):
+        self.read_from = read_from
+        self.write_to = write_to
+        self.calls = []
+
+    def connect_ssh(self, username, password, host, port, command):
+        self.calls.append(('connect_ssh', username, password, host, port,
+            command))
+        return StringIOSSHConnection(self)
+
+
+class StringIOSSHConnection(object):
+    """A SSH connection that uses StringIO to buffer writes and answer reads."""
+
+    def __init__(self, vendor):
+        self.vendor = vendor
+    
+    def close(self):
+        self.vendor.calls.append(('close', ))
+        
+    def get_filelike_channels(self):
+        return self.vendor.read_from, self.vendor.write_to
+
+
+
+class SmartClientMediumTests(tests.TestCase):
+    """Tests for SmartClientMedium.
+
+    We should create a test scenario for this: we need a server module that
+    construct the test-servers (like make_loopsocket_and_medium), and the list
+    of SmartClientMedium classes to test.
+    """
+
+    def make_loopsocket_and_medium(self):
+        """Create a loopback socket for testing, and a medium aimed at it."""
+        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        sock.bind(('127.0.0.1', 0))
+        sock.listen(1)
+        port = sock.getsockname()[1]
+        medium = smart.SmartTCPClientMedium('127.0.0.1', port)
+        return sock, medium
+
+    def receive_bytes_on_server(self, sock, bytes):
+        """Accept a connection on sock and read 3 bytes.
+
+        The bytes are appended to the list bytes.
+
+        :return: a Thread which is running to do the accept and recv.
+        """
+        def _receive_bytes_on_server():
+            connection, address = sock.accept()
+            bytes.append(connection.recv(3, socket.MSG_WAITALL))
+            connection.close()
+        t = threading.Thread(target=_receive_bytes_on_server)
+        t.start()
+        return t
+    
+    def test_construct_smart_stream_medium_client(self):
+        # make a new instance of the common base for Stream-like Mediums.
+        # this just ensures that the constructor stays parameter-free which
+        # is important for reuse : some subclasses will dynamically connect,
+        # others are always on, etc.
+        medium = smart.SmartClientStreamMedium()
+
+    def test_construct_smart_client_medium(self):
+        # the base client medium takes no parameters
+        medium = smart.SmartClientMedium()
+    
+    def test_construct_smart_simple_pipes_client_medium(self):
+        # the SimplePipes client medium takes two pipes:
+        # readable pipe, writeable pipe.
+        # Constructing one should just save these and do nothing.
+        # We test this by passing in None.
+        medium = smart.SmartSimplePipesClientMedium(None, None)
+        
+    def test_simple_pipes_client_request_type(self):
+        # SimplePipesClient should use SmartClientStreamMediumRequest's.
+        medium = smart.SmartSimplePipesClientMedium(None, None)
+        request = medium.get_request()
+        self.assertIsInstance(request, smart.SmartClientStreamMediumRequest)
+
+    def test_simple_pipes_client_get_concurrent_requests(self):
+        # the simple_pipes client does not support pipelined requests:
+        # but it does support serial requests: we construct one after 
+        # another is finished. This is a smoke test testing the integration
+        # of the SmartClientStreamMediumRequest and the SmartClientStreamMedium
+        # classes - as the sibling classes share this logic, they do not have
+        # explicit tests for this.
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = medium.get_request()
+        request.finished_writing()
+        request.finished_reading()
+        request2 = medium.get_request()
+        request2.finished_writing()
+        request2.finished_reading()
+
+    def test_simple_pipes_client__accept_bytes_writes_to_writable(self):
+        # accept_bytes writes to the writeable pipe.
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        medium._accept_bytes('abc')
+        self.assertEqual('abc', output.getvalue())
+    
+    def test_simple_pipes_client_disconnect_does_nothing(self):
+        # calling disconnect does nothing.
+        input = StringIO()
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        # send some bytes to ensure disconnecting after activity still does not
+        # close.
+        medium._accept_bytes('abc')
+        medium.disconnect()
+        self.assertFalse(input.closed)
+        self.assertFalse(output.closed)
+
+    def test_simple_pipes_client_accept_bytes_after_disconnect(self):
+        # calling disconnect on the client does not alter the pipe that
+        # accept_bytes writes to.
+        input = StringIO()
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        medium._accept_bytes('abc')
+        medium.disconnect()
+        medium._accept_bytes('abc')
+        self.assertFalse(input.closed)
+        self.assertFalse(output.closed)
+        self.assertEqual('abcabc', output.getvalue())
+    
+    def test_simple_pipes_client_ignores_disconnect_when_not_connected(self):
+        # Doing a disconnect on a new (and thus unconnected) SimplePipes medium
+        # does nothing.
+        medium = smart.SmartSimplePipesClientMedium(None, None)
+        medium.disconnect()
+
+    def test_simple_pipes_client_can_always_read(self):
+        # SmartSimplePipesClientMedium is never disconnected, so read_bytes
+        # always tries to read from the underlying pipe.
+        input = StringIO('abcdef')
+        medium = smart.SmartSimplePipesClientMedium(input, None)
+        self.assertEqual('abc', medium.read_bytes(3))
+        medium.disconnect()
+        self.assertEqual('def', medium.read_bytes(3))
+        
+    def test_simple_pipes_client_supports__flush(self):
+        # invoking _flush on a SimplePipesClient should flush the output 
+        # pipe. We test this by creating an output pipe that records
+        # flush calls made to it.
+        from StringIO import StringIO # get regular StringIO
+        input = StringIO()
+        output = StringIO()
+        flush_calls = []
+        def logging_flush(): flush_calls.append('flush')
+        output.flush = logging_flush
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        # this call is here to ensure we only flush once, not on every
+        # _accept_bytes call.
+        medium._accept_bytes('abc')
+        medium._flush()
+        medium.disconnect()
+        self.assertEqual(['flush'], flush_calls)
+
+    def test_construct_smart_ssh_client_medium(self):
+        # the SSH client medium takes:
+        # host, port, username, password, vendor
+        # Constructing one should just save these and do nothing.
+        # we test this by creating a empty bound socket and constructing
+        # a medium.
+        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        sock.bind(('127.0.0.1', 0))
+        unopened_port = sock.getsockname()[1]
+        # having vendor be invalid means that if it tries to connect via the
+        # vendor it will blow up.
+        medium = smart.SmartSSHClientMedium('127.0.0.1', unopened_port,
+            username=None, password=None, vendor="not a vendor")
+        sock.close()
+
+    def test_ssh_client_connects_on_first_use(self):
+        # The only thing that initiates a connection from the medium is giving
+        # it bytes.
+        output = StringIO()
+        vendor = StringIOSSHVendor(StringIO(), output)
+        medium = smart.SmartSSHClientMedium('a hostname', 'a port', 'a username',
+            'a password', vendor)
+        medium._accept_bytes('abc')
+        self.assertEqual('abc', output.getvalue())
+        self.assertEqual([('connect_ssh', 'a username', 'a password',
+            'a hostname', 'a port',
+            ['bzr', 'serve', '--inet', '--directory=/', '--allow-writes'])],
+            vendor.calls)
+    
+    def test_ssh_client_changes_command_when_BZR_REMOTE_PATH_is_set(self):
+        # The only thing that initiates a connection from the medium is giving
+        # it bytes.
+        output = StringIO()
+        vendor = StringIOSSHVendor(StringIO(), output)
+        orig_bzr_remote_path = os.environ.get('BZR_REMOTE_PATH')
+        def cleanup_environ():
+            osutils.set_or_unset_env('BZR_REMOTE_PATH', orig_bzr_remote_path)
+        self.addCleanup(cleanup_environ)
+        os.environ['BZR_REMOTE_PATH'] = 'fugly'
+        medium = smart.SmartSSHClientMedium('a hostname', 'a port', 'a username',
+            'a password', vendor)
+        medium._accept_bytes('abc')
+        self.assertEqual('abc', output.getvalue())
+        self.assertEqual([('connect_ssh', 'a username', 'a password',
+            'a hostname', 'a port',
+            ['fugly', 'serve', '--inet', '--directory=/', '--allow-writes'])],
+            vendor.calls)
+    
+    def test_ssh_client_disconnect_does_so(self):
+        # calling disconnect should disconnect both the read_from and write_to
+        # file-like object it from the ssh connection.
+        input = StringIO()
+        output = StringIO()
+        vendor = StringIOSSHVendor(input, output)
+        medium = smart.SmartSSHClientMedium('a hostname', vendor=vendor)
+        medium._accept_bytes('abc')
+        medium.disconnect()
+        self.assertTrue(input.closed)
+        self.assertTrue(output.closed)
+        self.assertEqual([
+            ('connect_ssh', None, None, 'a hostname', None,
+            ['bzr', 'serve', '--inet', '--directory=/', '--allow-writes']),
+            ('close', ),
+            ],
+            vendor.calls)
+
+    def test_ssh_client_disconnect_allows_reconnection(self):
+        # calling disconnect on the client terminates the connection, but should
+        # not prevent additional connections occuring.
+        # we test this by initiating a second connection after doing a
+        # disconnect.
+        input = StringIO()
+        output = StringIO()
+        vendor = StringIOSSHVendor(input, output)
+        medium = smart.SmartSSHClientMedium('a hostname', vendor=vendor)
+        medium._accept_bytes('abc')
+        medium.disconnect()
+        # the disconnect has closed output, so we need a new output for the
+        # new connection to write to.
+        input2 = StringIO()
+        output2 = StringIO()
+        vendor.read_from = input2
+        vendor.write_to = output2
+        medium._accept_bytes('abc')
+        medium.disconnect()
+        self.assertTrue(input.closed)
+        self.assertTrue(output.closed)
+        self.assertTrue(input2.closed)
+        self.assertTrue(output2.closed)
+        self.assertEqual([
+            ('connect_ssh', None, None, 'a hostname', None,
+            ['bzr', 'serve', '--inet', '--directory=/', '--allow-writes']),
+            ('close', ),
+            ('connect_ssh', None, None, 'a hostname', None,
+            ['bzr', 'serve', '--inet', '--directory=/', '--allow-writes']),
+            ('close', ),
+            ],
+            vendor.calls)
+    
+    def test_ssh_client_ignores_disconnect_when_not_connected(self):
+        # Doing a disconnect on a new (and thus unconnected) SSH medium
+        # does not fail.  It's ok to disconnect an unconnected medium.
+        medium = smart.SmartSSHClientMedium(None)
+        medium.disconnect()
+
+    def test_ssh_client_raises_on_read_when_not_connected(self):
+        # Doing a read on a new (and thus unconnected) SSH medium raises
+        # MediumNotConnected.
+        medium = smart.SmartSSHClientMedium(None)
+        self.assertRaises(errors.MediumNotConnected, medium.read_bytes, 0)
+        self.assertRaises(errors.MediumNotConnected, medium.read_bytes, 1)
+
+    def test_ssh_client_supports__flush(self):
+        # invoking _flush on a SSHClientMedium should flush the output 
+        # pipe. We test this by creating an output pipe that records
+        # flush calls made to it.
+        from StringIO import StringIO # get regular StringIO
+        input = StringIO()
+        output = StringIO()
+        flush_calls = []
+        def logging_flush(): flush_calls.append('flush')
+        output.flush = logging_flush
+        vendor = StringIOSSHVendor(input, output)
+        medium = smart.SmartSSHClientMedium('a hostname', vendor=vendor)
+        # this call is here to ensure we only flush once, not on every
+        # _accept_bytes call.
+        medium._accept_bytes('abc')
+        medium._flush()
+        medium.disconnect()
+        self.assertEqual(['flush'], flush_calls)
+        
+    def test_construct_smart_tcp_client_medium(self):
+        # the TCP client medium takes a host and a port.  Constructing it won't
+        # connect to anything.
+        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        sock.bind(('127.0.0.1', 0))
+        unopened_port = sock.getsockname()[1]
+        medium = smart.SmartTCPClientMedium('127.0.0.1', unopened_port)
+        sock.close()
+
+    def test_tcp_client_connects_on_first_use(self):
+        # The only thing that initiates a connection from the medium is giving
+        # it bytes.
+        sock, medium = self.make_loopsocket_and_medium()
+        bytes = []
+        t = self.receive_bytes_on_server(sock, bytes)
+        medium.accept_bytes('abc')
+        t.join()
+        sock.close()
+        self.assertEqual(['abc'], bytes)
+    
+    def test_tcp_client_disconnect_does_so(self):
+        # calling disconnect on the client terminates the connection.
+        # we test this by forcing a short read during a socket.MSG_WAITALL
+        # call: write 2 bytes, try to read 3, and then the client disconnects.
+        sock, medium = self.make_loopsocket_and_medium()
+        bytes = []
+        t = self.receive_bytes_on_server(sock, bytes)
+        medium.accept_bytes('ab')
+        medium.disconnect()
+        t.join()
+        sock.close()
+        self.assertEqual(['ab'], bytes)
+        # now disconnect again: this should not do anything, if disconnection
+        # really did disconnect.
+        medium.disconnect()
+    
+    def test_tcp_client_ignores_disconnect_when_not_connected(self):
+        # Doing a disconnect on a new (and thus unconnected) TCP medium
+        # does not fail.  It's ok to disconnect an unconnected medium.
+        medium = smart.SmartTCPClientMedium(None, None)
+        medium.disconnect()
+
+    def test_tcp_client_raises_on_read_when_not_connected(self):
+        # Doing a read on a new (and thus unconnected) TCP medium raises
+        # MediumNotConnected.
+        medium = smart.SmartTCPClientMedium(None, None)
+        self.assertRaises(errors.MediumNotConnected, medium.read_bytes, 0)
+        self.assertRaises(errors.MediumNotConnected, medium.read_bytes, 1)
+
+    def test_tcp_client_supports__flush(self):
+        # invoking _flush on a TCPClientMedium should do something useful.
+        # RBC 20060922 not sure how to test/tell in this case.
+        sock, medium = self.make_loopsocket_and_medium()
+        bytes = []
+        t = self.receive_bytes_on_server(sock, bytes)
+        # try with nothing buffered
+        medium._flush()
+        medium._accept_bytes('ab')
+        # and with something sent.
+        medium._flush()
+        medium.disconnect()
+        t.join()
+        sock.close()
+        self.assertEqual(['ab'], bytes)
+        # now disconnect again : this should not do anything, if disconnection
+        # really did disconnect.
+        medium.disconnect()
+
+
+class TestSmartClientStreamMediumRequest(tests.TestCase):
+    """Tests the for SmartClientStreamMediumRequest.
+    
+    SmartClientStreamMediumRequest is a helper for the three stream based 
+    mediums: TCP, SSH, SimplePipes, so we only test it once, and then test that
+    those three mediums implement the interface it expects.
+    """
+
+    def test_accept_bytes_after_finished_writing_errors(self):
+        # calling accept_bytes after calling finished_writing raises 
+        # WritingCompleted to prevent bad assumptions on stream environments
+        # breaking the needs of message-based environments.
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.finished_writing()
+        self.assertRaises(errors.WritingCompleted, request.accept_bytes, None)
+
+    def test_accept_bytes(self):
+        # accept bytes should invoke _accept_bytes on the stream medium.
+        # we test this by using the SimplePipes medium - the most trivial one
+        # and checking that the pipes get the data.
+        input = StringIO()
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.accept_bytes('123')
+        request.finished_writing()
+        request.finished_reading()
+        self.assertEqual('', input.getvalue())
+        self.assertEqual('123', output.getvalue())
+
+    def test_construct_sets_stream_request(self):
+        # constructing a SmartClientStreamMediumRequest on a StreamMedium sets
+        # the current request to the new SmartClientStreamMediumRequest
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        self.assertIs(medium._current_request, request)
+
+    def test_construct_while_another_request_active_throws(self):
+        # constructing a SmartClientStreamMediumRequest on a StreamMedium with
+        # a non-None _current_request raises TooManyConcurrentRequests.
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        medium._current_request = "a"
+        self.assertRaises(errors.TooManyConcurrentRequests,
+            smart.SmartClientStreamMediumRequest, medium)
+
+    def test_finished_read_clears_current_request(self):
+        # calling finished_reading clears the current request from the requests
+        # medium
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.finished_writing()
+        request.finished_reading()
+        self.assertEqual(None, medium._current_request)
+
+    def test_finished_read_before_finished_write_errors(self):
+        # calling finished_reading before calling finished_writing triggers a
+        # WritingNotComplete error.
+        medium = smart.SmartSimplePipesClientMedium(None, None)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        self.assertRaises(errors.WritingNotComplete, request.finished_reading)
+        
+    def test_read_bytes(self):
+        # read bytes should invoke _read_bytes on the stream medium.
+        # we test this by using the SimplePipes medium - the most trivial one
+        # and checking that the data is supplied. Its possible that a 
+        # faulty implementation could poke at the pipe variables them selves,
+        # but we trust that this will be caught as it will break the integration
+        # smoke tests.
+        input = StringIO('321')
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.finished_writing()
+        self.assertEqual('321', request.read_bytes(3))
+        request.finished_reading()
+        self.assertEqual('', input.read())
+        self.assertEqual('', output.getvalue())
+
+    def test_read_bytes_before_finished_write_errors(self):
+        # calling read_bytes before calling finished_writing triggers a
+        # WritingNotComplete error because the Smart protocol is designed to be
+        # compatible with strict message based protocols like HTTP where the
+        # request cannot be submitted until the writing has completed.
+        medium = smart.SmartSimplePipesClientMedium(None, None)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        self.assertRaises(errors.WritingNotComplete, request.read_bytes, None)
+
+    def test_read_bytes_after_finished_reading_errors(self):
+        # calling read_bytes after calling finished_reading raises 
+        # ReadingCompleted to prevent bad assumptions on stream environments
+        # breaking the needs of message-based environments.
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = smart.SmartClientStreamMediumRequest(medium)
+        request.finished_writing()
+        request.finished_reading()
+        self.assertRaises(errors.ReadingCompleted, request.read_bytes, None)
+
+
+class RemoteTransportTests(tests.TestCaseWithTransport):
 
     def setUp(self):
-        super(TCPClientTests, self).setUp()
+        super(RemoteTransportTests, self).setUp()
         # We're allowed to set  the transport class here, so that we don't use
         # the default or a parameterized class, but rather use the
         # TestCaseWithTransport infrastructure to set up a smart server and
@@ -61,21 +529,70 @@
         t = self.get_transport()
         self.assertIsInstance(t, smart.SmartTransport)
 
-    def test_get_client_from_transport(self):
+    def test_get_medium_from_transport(self):
+        """Remote transport has a medium always, which it can return."""
         t = self.get_transport()
-        client = t.get_smart_client()
-        self.assertIsInstance(client, smart.SmartStreamClient)
-
-
-class BasicSmartTests(tests.TestCase):
+        medium = t.get_smart_medium()
+        self.assertIsInstance(medium, smart.SmartClientMedium)
+
+
+class ErrorRaisingProtocol(object):
+
+    def __init__(self, exception):
+        self.exception = exception
+
+    def next_read_size(self):
+        raise self.exception
+
+
+class SampleRequest(object):
+    
+    def __init__(self, expected_bytes):
+        self.accepted_bytes = ''
+        self._finished_reading = False
+        self.expected_bytes = expected_bytes
+        self.excess_buffer = ''
+
+    def accept_bytes(self, bytes):
+        self.accepted_bytes += bytes
+        if self.accepted_bytes.startswith(self.expected_bytes):
+            self._finished_reading = True
+            self.excess_buffer = self.accepted_bytes[len(self.expected_bytes):]
+
+    def next_read_size(self):
+        if self._finished_reading:
+            return 0
+        else:
+            return 1
+
+
+class TestSmartServerStreamMedium(tests.TestCase):
+
+    def portable_socket_pair(self):
+        """Return a pair of TCP sockets connected to each other.
+        
+        Unlike socket.socketpair, this should work on Windows.
+        """
+        listen_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        listen_sock.bind(('127.0.0.1', 0))
+        listen_sock.listen(1)
+        client_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        client_sock.connect(listen_sock.getsockname())
+        server_sock, addr = listen_sock.accept()
+        listen_sock.close()
+        return server_sock, client_sock
     
     def test_smart_query_version(self):
         """Feed a canned query version to a server"""
+        # wire-to-wire, using the whole stack
         to_server = StringIO('hello\n')
         from_server = StringIO()
-        server = smart.SmartStreamServer(to_server, from_server, 
-            local.LocalTransport(urlutils.local_path_to_url('/')))
-        server._serve_one_request()
+        transport = local.LocalTransport(urlutils.local_path_to_url('/'))
+        server = smart.SmartServerPipeStreamMedium(
+            to_server, from_server, transport)
+        protocol = smart.SmartServerRequestProtocolOne(transport,
+                from_server.write)
+        server._serve_one_request(protocol)
         self.assertEqual('ok\0011\n',
                          from_server.getvalue())
 
@@ -84,14 +601,155 @@
         transport.put_bytes('testfile', 'contents\nof\nfile\n')
         to_server = StringIO('get\001./testfile\n')
         from_server = StringIO()
-        server = smart.SmartStreamServer(to_server, from_server, transport)
-        server._serve_one_request()
+        server = smart.SmartServerPipeStreamMedium(
+            to_server, from_server, transport)
+        protocol = smart.SmartServerRequestProtocolOne(transport,
+                from_server.write)
+        server._serve_one_request(protocol)
         self.assertEqual('ok\n'
                          '17\n'
                          'contents\nof\nfile\n'
                          'done\n',
                          from_server.getvalue())
 
+    def test_pipe_like_stream_with_bulk_data(self):
+        sample_request_bytes = 'command\n9\nbulk datadone\n'
+        to_server = StringIO(sample_request_bytes)
+        from_server = StringIO()
+        server = smart.SmartServerPipeStreamMedium(to_server, from_server, None)
+        sample_protocol = SampleRequest(expected_bytes=sample_request_bytes)
+        server._serve_one_request(sample_protocol)
+        self.assertEqual('', from_server.getvalue())
+        self.assertEqual(sample_request_bytes, sample_protocol.accepted_bytes)
+        self.assertFalse(server.finished)
+
+    def test_socket_stream_with_bulk_data(self):
+        sample_request_bytes = 'command\n9\nbulk datadone\n'
+        server_sock, client_sock = self.portable_socket_pair()
+        server = smart.SmartServerSocketStreamMedium(
+            server_sock, None)
+        sample_protocol = SampleRequest(expected_bytes=sample_request_bytes)
+        client_sock.sendall(sample_request_bytes)
+        server._serve_one_request(sample_protocol)
+        server_sock.close()
+        self.assertEqual('', client_sock.recv(1))
+        self.assertEqual(sample_request_bytes, sample_protocol.accepted_bytes)
+        self.assertFalse(server.finished)
+
+    def test_pipe_like_stream_shutdown_detection(self):
+        to_server = StringIO('')
+        from_server = StringIO()
+        server = smart.SmartServerPipeStreamMedium(to_server, from_server, None)
+        server._serve_one_request(SampleRequest('x'))
+        self.assertTrue(server.finished)
+        
+    def test_socket_stream_shutdown_detection(self):
+        server_sock, client_sock = self.portable_socket_pair()
+        client_sock.close()
+        server = smart.SmartServerSocketStreamMedium(
+            server_sock, None)
+        server._serve_one_request(SampleRequest('x'))
+        self.assertTrue(server.finished)
+        
+    def test_pipe_like_stream_with_two_requests(self):
+        # If two requests are read in one go, then two calls to
+        # _serve_one_request should still process both of them as if they had
+        # been received seperately.
+        sample_request_bytes = 'command\n'
+        to_server = StringIO(sample_request_bytes * 2)
+        from_server = StringIO()
+        server = smart.SmartServerPipeStreamMedium(to_server, from_server, None)
+        first_protocol = SampleRequest(expected_bytes=sample_request_bytes)
+        server._serve_one_request(first_protocol)
+        self.assertEqual(0, first_protocol.next_read_size())
+        self.assertEqual('', from_server.getvalue())
+        self.assertFalse(server.finished)
+        # Make a new protocol, call _serve_one_request with it to collect the
+        # second request.
+        second_protocol = SampleRequest(expected_bytes=sample_request_bytes)
+        server._serve_one_request(second_protocol)
+        self.assertEqual('', from_server.getvalue())
+        self.assertEqual(sample_request_bytes, second_protocol.accepted_bytes)
+        self.assertFalse(server.finished)
+        
+    def test_socket_stream_with_two_requests(self):
+        # If two requests are read in one go, then two calls to
+        # _serve_one_request should still process both of them as if they had
+        # been received seperately.
+        sample_request_bytes = 'command\n'
+        server_sock, client_sock = self.portable_socket_pair()
+        server = smart.SmartServerSocketStreamMedium(
+            server_sock, None)
+        first_protocol = SampleRequest(expected_bytes=sample_request_bytes)
+        # Put two whole requests on the wire.
+        client_sock.sendall(sample_request_bytes * 2)
+        server._serve_one_request(first_protocol)
+        self.assertEqual(0, first_protocol.next_read_size())
+        self.assertFalse(server.finished)
+        # Make a new protocol, call _serve_one_request with it to collect the
+        # second request.
+        second_protocol = SampleRequest(expected_bytes=sample_request_bytes)
+        stream_still_open = server._serve_one_request(second_protocol)
+        self.assertEqual(sample_request_bytes, second_protocol.accepted_bytes)
+        self.assertFalse(server.finished)
+        server_sock.close()
+        self.assertEqual('', client_sock.recv(1))
+
+    def test_pipe_like_stream_error_handling(self):
+        # Use plain python StringIO so we can monkey-patch the close method to
+        # not discard the contents.
+        from StringIO import StringIO
+        to_server = StringIO('')
+        from_server = StringIO()
+        self.closed = False
+        def close():
+            self.closed = True
+        from_server.close = close
+        server = smart.SmartServerPipeStreamMedium(to_server, from_server, None)
+        fake_protocol = ErrorRaisingProtocol(Exception('boom'))
+        server._serve_one_request(fake_protocol)
+        self.assertEqual('', from_server.getvalue())
+        self.assertTrue(self.closed)
+        self.assertTrue(server.finished)
+        
+    def test_socket_stream_error_handling(self):
+        # Use plain python StringIO so we can monkey-patch the close method to
+        # not discard the contents.
+        from StringIO import StringIO
+        server_sock, client_sock = self.portable_socket_pair()
+        server = smart.SmartServerSocketStreamMedium(
+            server_sock, None)
+        fake_protocol = ErrorRaisingProtocol(Exception('boom'))
+        server._serve_one_request(fake_protocol)
+        # recv should not block, because the other end of the socket has been
+        # closed.
+        self.assertEqual('', client_sock.recv(1))
+        self.assertTrue(server.finished)
+        
+    def test_pipe_like_stream_keyboard_interrupt_handling(self):
+        # Use plain python StringIO so we can monkey-patch the close method to
+        # not discard the contents.
+        to_server = StringIO('')
+        from_server = StringIO()
+        server = smart.SmartServerPipeStreamMedium(to_server, from_server, None)
+        fake_protocol = ErrorRaisingProtocol(KeyboardInterrupt('boom'))
+        self.assertRaises(
+            KeyboardInterrupt, server._serve_one_request, fake_protocol)
+        self.assertEqual('', from_server.getvalue())
+
+    def test_socket_stream_keyboard_interrupt_handling(self):
+        server_sock, client_sock = self.portable_socket_pair()
+        server = smart.SmartServerSocketStreamMedium(
+            server_sock, None)
+        fake_protocol = ErrorRaisingProtocol(KeyboardInterrupt('boom'))
+        self.assertRaises(
+            KeyboardInterrupt, server._serve_one_request, fake_protocol)
+        server_sock.close()
+        self.assertEqual('', client_sock.recv(1))
+        
+
+class TestSmartTCPServer(tests.TestCase):
+
     def test_get_error_unexpected(self):
         """Error reported by server with no specific representation"""
         class FlakyTransport(object):
@@ -110,25 +768,6 @@
         finally:
             server.stop_background_thread()
 
-    def test_server_subprocess(self):
-        """Talk to a server started as a subprocess
-        
-        This is similar to running it over ssh, except that it runs in the same machine 
-        without ssh intermediating.
-        """
-        args = [sys.executable, sys.argv[0], 'serve', '--inet']
-        do_close_fds = True
-        if sys.platform == 'win32':
-            do_close_fds = False
-        child = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
-                                 close_fds=do_close_fds, universal_newlines=True)
-        conn = smart.SmartStreamClient(lambda: (child.stdout, child.stdin))
-        conn.query_version()
-        conn.query_version()
-        conn.disconnect()
-        returncode = child.wait()
-        self.assertEquals(0, returncode)
-
 
 class SmartTCPTests(tests.TestCase):
     """Tests for connection/end to end behaviour using the TCP server.
@@ -200,7 +839,7 @@
         # we create a real connection not a loopback one, but it will use the
         # same server and pipes
         conn2 = self.transport.clone('.')
-        self.assertTrue(self.transport._client is conn2._client)
+        self.assertIs(self.transport._medium, conn2._medium)
 
     def test__remote_path(self):
         self.assertEquals('/foo/bar',
@@ -246,14 +885,19 @@
             'foo')
         
 
-class SmartServerTests(tests.TestCaseWithTransport):
-    """Test that call directly into the server logic, bypassing the network."""
+class SmartServerRequestHandlerTests(tests.TestCaseWithTransport):
+    """Test that call directly into the handler logic, bypassing the network."""
+
+    def test_construct_request_handler(self):
+        """Constructing a request handler should be easy and set defaults."""
+        handler = smart.SmartServerRequestHandler(None)
+        self.assertFalse(handler.finished_reading)
 
     def test_hello(self):
-        server = smart.SmartServer(None)
-        response = server.dispatch_command('hello', ())
-        self.assertEqual(('ok', '1'), response.args)
-        self.assertEqual(None, response.body)
+        handler = smart.SmartServerRequestHandler(None)
+        handler.dispatch_command('hello', ())
+        self.assertEqual(('ok', '1'), handler.response.args)
+        self.assertEqual(None, handler.response.body)
         
     def test_get_bundle(self):
         from bzrlib.bundle import serializer
@@ -262,22 +906,71 @@
         wt.add('hello')
         rev_id = wt.commit('add hello')
         
-        server = smart.SmartServer(self.get_transport())
-        response = server.dispatch_command('get_bundle', ('.', rev_id))
-        bundle = serializer.read_bundle(StringIO(response.body))
+        handler = smart.SmartServerRequestHandler(self.get_transport())
+        handler.dispatch_command('get_bundle', ('.', rev_id))
+        bundle = serializer.read_bundle(StringIO(handler.response.body))
+        self.assertEqual((), handler.response.args)
 
     def test_readonly_exception_becomes_transport_not_possible(self):
         """The response for a read-only error is ('ReadOnlyError')."""
-        server = smart.SmartServer(self.get_readonly_transport())
+        handler = smart.SmartServerRequestHandler(self.get_readonly_transport())
         # send a mkdir for foo, with no explicit mode - should fail.
-        response = server.dispatch_command('mkdir', ('foo', ''))
+        handler.dispatch_command('mkdir', ('foo', ''))
         # and the failure should be an explicit ReadOnlyError
-        self.assertEqual(("ReadOnlyError", ), response.args)
+        self.assertEqual(("ReadOnlyError", ), handler.response.args)
         # XXX: TODO: test that other TransportNotPossible errors are
         # presented as TransportNotPossible - not possible to do that
         # until I figure out how to trigger that relatively cleanly via
         # the api. RBC 20060918
 
+    def test_hello_has_finished_body_on_dispatch(self):
+        """The 'hello' command should set finished_reading."""
+        handler = smart.SmartServerRequestHandler(None)
+        handler.dispatch_command('hello', ())
+        self.assertTrue(handler.finished_reading)
+        self.assertNotEqual(None, handler.response)
+
+    def test_put_bytes_non_atomic(self):
+        """'put_...' should set finished_reading after reading the bytes."""
+        handler = smart.SmartServerRequestHandler(self.get_transport())
+        handler.dispatch_command('put_non_atomic', ('a-file', '', 'F', ''))
+        self.assertFalse(handler.finished_reading)
+        handler.accept_body('1234')
+        self.assertFalse(handler.finished_reading)
+        handler.accept_body('5678')
+        handler.end_of_body()
+        self.assertTrue(handler.finished_reading)
+        self.assertEqual(('ok', ), handler.response.args)
+        self.assertEqual(None, handler.response.body)
+        
+    def test_readv_accept_body(self):
+        """'readv' should set finished_reading after reading offsets."""
+        self.build_tree(['a-file'])
+        handler = smart.SmartServerRequestHandler(self.get_readonly_transport())
+        handler.dispatch_command('readv', ('a-file', ))
+        self.assertFalse(handler.finished_reading)
+        handler.accept_body('2,')
+        self.assertFalse(handler.finished_reading)
+        handler.accept_body('3')
+        handler.end_of_body()
+        self.assertTrue(handler.finished_reading)
+        self.assertEqual(('readv', ), handler.response.args)
+        # co - nte - nt of a-file is the file contents we are extracting from.
+        self.assertEqual('nte', handler.response.body)
+
+    def test_readv_short_read_response_contents(self):
+        """'readv' when a short read occurs sets the response appropriately."""
+        self.build_tree(['a-file'])
+        handler = smart.SmartServerRequestHandler(self.get_readonly_transport())
+        handler.dispatch_command('readv', ('a-file', ))
+        # read beyond the end of the file.
+        handler.accept_body('100,1')
+        handler.end_of_body()
+        self.assertTrue(handler.finished_reading)
+        self.assertEqual(('ShortReadvError', 'a-file', '100', '1', '0'),
+            handler.response.args)
+        self.assertEqual(None, handler.response.body)
+
 
 class SmartTransportRegistration(tests.TestCase):
 
@@ -287,70 +980,44 @@
         self.assertEqual('example.com', t._host)
 
 
-class FakeClient(smart.SmartStreamClient):
-    """Emulate a client for testing a transport's use of the client."""
-
-    def __init__(self):
-        smart.SmartStreamClient.__init__(self, None)
-        self._calls = []
-
-    def _call(self, *args):
-        self._calls.append(('_call', args))
-        return ('ok', )
-
-    def _recv_bulk(self):
-        return 'bar'
-
-
 class TestSmartTransport(tests.TestCase):
         
     def test_use_connection_factory(self):
         # We want to be able to pass a client as a parameter to SmartTransport.
-        client = FakeClient()
-        transport = smart.SmartTransport('bzr://localhost/', client=client)
+        input = StringIO("ok\n3\nbardone\n")
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        transport = smart.SmartTransport('bzr://localhost/', medium=medium)
 
         # We want to make sure the client is used when the first remote
-        # method is called.  No method should have been called yet.
-        self.assertEqual([], client._calls)
+        # method is called.  No data should have been sent, or read.
+        self.assertEqual(0, input.tell())
+        self.assertEqual('', output.getvalue())
 
-        # Now call a method that should result in a single request.
+        # Now call a method that should result in a single request : as the
+        # transport makes its own protocol instances, we check on the wire.
+        # XXX: TODO: give the transport a protocol factory, which can make
+        # an instrumented protocol for us.
         self.assertEqual('bar', transport.get_bytes('foo'))
-        # The only call to _call should have been to get /foo.
-        self.assertEqual([('_call', ('get', '/foo'))], client._calls)
+        # only the needed data should have been sent/received.
+        self.assertEqual(13, input.tell())
+        self.assertEqual('get\x01/foo\n', output.getvalue())
 
     def test__translate_error_readonly(self):
         """Sending a ReadOnlyError to _translate_error raises TransportNotPossible."""
-        client = FakeClient()
-        transport = smart.SmartTransport('bzr://localhost/', client=client)
+        medium = smart.SmartClientMedium()
+        transport = smart.SmartTransport('bzr://localhost/', medium=medium)
         self.assertRaises(errors.TransportNotPossible,
             transport._translate_error, ("ReadOnlyError", ))
 
 
-class InstrumentedClient(smart.SmartStreamClient):
-    """A smart client whose writes are stored to a supplied list."""
-
-    def __init__(self, write_output_list):
-        smart.SmartStreamClient.__init__(self, None)
-        self._write_output_list = write_output_list
-
-    def _ensure_connection(self):
-        """We are never strictly connected."""
-
-    def _write_and_flush(self, bytes):
-        self._write_output_list.append(bytes)
-
-
-class InstrumentedServerProtocol(smart.SmartStreamServer):
+class InstrumentedServerProtocol(smart.SmartServerStreamMedium):
     """A smart server which is backed by memory and saves its write requests."""
 
     def __init__(self, write_output_list):
-        smart.SmartStreamServer.__init__(self, None, None,
-            memory.MemoryTransport())
+        smart.SmartServerStreamMedium.__init__(self, memory.MemoryTransport())
         self._write_output_list = write_output_list
 
-    def _write_and_flush(self, bytes):
-        self._write_output_list.append(bytes)
-
 
 class TestSmartProtocol(tests.TestCase):
     """Tests for the smart protocol.
@@ -367,13 +1034,18 @@
 
     def setUp(self):
         super(TestSmartProtocol, self).setUp()
-        self.to_server = []
-        self.to_client = []
-        self.smart_client = InstrumentedClient(self.to_server)
-        self.smart_server = InstrumentedServerProtocol(self.to_client)
+        self.server_to_client = []
+        self.to_server = StringIO()
+        self.to_client = StringIO()
+        self.client_medium = smart.SmartSimplePipesClientMedium(self.to_client,
+            self.to_server)
+        self.client_protocol = smart.SmartClientRequestProtocolOne(
+            self.client_medium)
+        self.smart_server = InstrumentedServerProtocol(self.server_to_client)
+        self.smart_server_request = smart.SmartServerRequestHandler(None)
 
     def assertOffsetSerialisation(self, expected_offsets, expected_serialised,
-        client, server_protocol):
+        client, smart_server_request):
         """Check that smart (de)serialises offsets as expected.
         
         We check both serialisation and deserialisation at the same time
@@ -382,14 +1054,46 @@
         
         :param expected_offsets: a readv offset list.
         :param expected_seralised: an expected serial form of the offsets.
-        :param server: a SmartServer instance.
+        :param smart_server_request: a SmartServerRequestHandler instance.
         """
-        offsets = server_protocol.smart_server._deserialise_offsets(
-            expected_serialised)
+        # XXX: 'smart_server_request' should be a SmartServerRequestProtocol in
+        # future.
+        offsets = smart_server_request._deserialise_offsets(expected_serialised)
         self.assertEqual(expected_offsets, offsets)
         serialised = client._serialise_offsets(offsets)
         self.assertEqual(expected_serialised, serialised)
 
+    def build_protocol_waiting_for_body(self):
+        out_stream = StringIO()
+        protocol = smart.SmartServerRequestProtocolOne(None, out_stream.write)
+        protocol.has_dispatched = True
+        protocol.request = smart.SmartServerRequestHandler(None)
+        def handle_end_of_bytes():
+            self.end_received = True
+            self.assertEqual('abcdefg', protocol.request._body_bytes)
+            protocol.request.response = smart.SmartServerResponse(('ok', ))
+        protocol.request._end_of_body_handler = handle_end_of_bytes
+        # Call accept_bytes to make sure that internal state like _body_decoder
+        # is initialised.  This test should probably be given a clearer
+        # interface to work with that will not cause this inconsistency.
+        #   -- Andrew Bennetts, 2006-09-28
+        protocol.accept_bytes('')
+        return protocol
+
+    def test_construct_version_one_server_protocol(self):
+        protocol = smart.SmartServerRequestProtocolOne(None, None)
+        self.assertEqual('', protocol.excess_buffer)
+        self.assertEqual('', protocol.in_buffer)
+        self.assertFalse(protocol.has_dispatched)
+        self.assertEqual(1, protocol.next_read_size())
+
+    def test_construct_version_one_client_protocol(self):
+        # we can construct a client protocol from a client medium request
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(None, output)
+        request = medium.get_request()
+        client_protocol = smart.SmartClientRequestProtocolOne(request)
+
     def test_server_offset_serialisation(self):
         """The Smart protocol serialises offsets as a comma and \n string.
 
@@ -398,15 +1102,410 @@
         one that should coalesce.
         """
         self.assertOffsetSerialisation([], '',
-            self.smart_client, self.smart_server)
+            self.client_protocol, self.smart_server_request)
         self.assertOffsetSerialisation([(1,2)], '1,2',
-            self.smart_client, self.smart_server)
+            self.client_protocol, self.smart_server_request)
         self.assertOffsetSerialisation([(10,40), (0,5)], '10,40\n0,5',
-            self.smart_client, self.smart_server)
+            self.client_protocol, self.smart_server_request)
         self.assertOffsetSerialisation([(1,2), (3,4), (100, 200)],
-            '1,2\n3,4\n100,200', self.smart_client, self.smart_server)
-
-
+            '1,2\n3,4\n100,200', self.client_protocol, self.smart_server_request)
+
+    def test_accept_bytes_to_protocol(self):
+        out_stream = StringIO()
+        protocol = smart.SmartServerRequestProtocolOne(None, out_stream.write)
+        protocol.accept_bytes('abc')
+        self.assertEqual('abc', protocol.in_buffer)
+        protocol.accept_bytes('\n')
+        self.assertEqual("error\x01Generic bzr smart protocol error: bad request"
+            " u'abc'\n", out_stream.getvalue())
+        self.assertTrue(protocol.has_dispatched)
+        self.assertEqual(1, protocol.next_read_size())
+
+    def test_accept_bytes_with_invalid_utf8_to_protocol(self):
+        out_stream = StringIO()
+        protocol = smart.SmartServerRequestProtocolOne(None, out_stream.write)
+        # the byte 0xdd is not a valid UTF-8 string.
+        protocol.accept_bytes('\xdd\n')
+        self.assertEqual(
+            "error\x01Generic bzr smart protocol error: "
+            "one or more arguments of request '\\xdd\\n' are not valid UTF-8\n",
+            out_stream.getvalue())
+        self.assertTrue(protocol.has_dispatched)
+        self.assertEqual(1, protocol.next_read_size())
+
+    def test_accept_body_bytes_to_protocol(self):
+        protocol = self.build_protocol_waiting_for_body()
+        self.assertEqual(6, protocol.next_read_size())
+        protocol.accept_bytes('7\nabc')
+        self.assertEqual(9, protocol.next_read_size())
+        protocol.accept_bytes('defgd')
+        protocol.accept_bytes('one\n')
+        self.assertEqual(0, protocol.next_read_size())
+        self.assertTrue(self.end_received)
+
+    def test_accept_request_and_body_all_at_once(self):
+        mem_transport = memory.MemoryTransport()
+        mem_transport.put_bytes('foo', 'abcdefghij')
+        out_stream = StringIO()
+        protocol = smart.SmartServerRequestProtocolOne(mem_transport,
+                out_stream.write)
+        protocol.accept_bytes('readv\x01foo\n3\n3,3done\n')
+        self.assertEqual(0, protocol.next_read_size())
+        self.assertEqual('readv\n3\ndefdone\n', out_stream.getvalue())
+        self.assertEqual('', protocol.excess_buffer)
+        self.assertEqual('', protocol.in_buffer)
+
+    def test_accept_excess_bytes_are_preserved(self):
+        out_stream = StringIO()
+        protocol = smart.SmartServerRequestProtocolOne(None, out_stream.write)
+        protocol.accept_bytes('hello\nhello\n')
+        self.assertEqual("ok\x011\n", out_stream.getvalue())
+        self.assertEqual("hello\n", protocol.excess_buffer)
+        self.assertEqual("", protocol.in_buffer)
+
+    def test_accept_excess_bytes_after_body(self):
+        protocol = self.build_protocol_waiting_for_body()
+        protocol.accept_bytes('7\nabcdefgdone\nX')
+        self.assertTrue(self.end_received)
+        self.assertEqual("X", protocol.excess_buffer)
+        self.assertEqual("", protocol.in_buffer)
+        protocol.accept_bytes('Y')
+        self.assertEqual("XY", protocol.excess_buffer)
+        self.assertEqual("", protocol.in_buffer)
+
+    def test_accept_excess_bytes_after_dispatch(self):
+        out_stream = StringIO()
+        protocol = smart.SmartServerRequestProtocolOne(None, out_stream.write)
+        protocol.accept_bytes('hello\n')
+        self.assertEqual("ok\x011\n", out_stream.getvalue())
+        protocol.accept_bytes('hel')
+        self.assertEqual("hel", protocol.excess_buffer)
+        protocol.accept_bytes('lo\n')
+        self.assertEqual("hello\n", protocol.excess_buffer)
+        self.assertEqual("", protocol.in_buffer)
+
+    def test_sync_with_request_sets_finished_reading(self):
+        protocol = smart.SmartServerRequestProtocolOne(None, None)
+        request = smart.SmartServerRequestHandler(None)
+        protocol.sync_with_request(request)
+        self.assertEqual(1, protocol.next_read_size())
+        request.finished_reading = True
+        protocol.sync_with_request(request)
+        self.assertEqual(0, protocol.next_read_size())
+
+    def test_query_version(self):
+        """query_version on a SmartClientProtocolOne should return a number.
+        
+        The protocol provides the query_version because the domain level clients
+        may all need to be able to probe for capabilities.
+        """
+        # What we really want to test here is that SmartClientProtocolOne calls
+        # accept_bytes(tuple_based_encoding_of_hello) and reads and parses the
+        # response of tuple-encoded (ok, 1).  Also, seperately we should test
+        # the error if the response is a non-understood version.
+        input = StringIO('ok\x011\n')
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        self.assertEqual(1, protocol.query_version())
+
+    def assertServerToClientEncoding(self, expected_bytes, expected_tuple,
+            input_tuples):
+        """Assert that each input_tuple serialises as expected_bytes, and the
+        bytes deserialise as expected_tuple.
+        """
+        # check the encoding of the server for all input_tuples matches
+        # expected bytes
+        for input_tuple in input_tuples:
+            server_output = StringIO()
+            server_protocol = smart.SmartServerRequestProtocolOne(
+                None, server_output.write)
+            server_protocol._send_response(input_tuple)
+            self.assertEqual(expected_bytes, server_output.getvalue())
+        # check the decoding of the client protocol from expected_bytes:
+        input = StringIO(expected_bytes)
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call('foo')
+        self.assertEqual(expected_tuple, protocol.read_response_tuple())
+
+    def test_client_call_empty_response(self):
+        # protocol.call() can get back an empty tuple as a response. This occurs
+        # when the parsed line is an empty line, and results in a tuple with
+        # one element - an empty string.
+        self.assertServerToClientEncoding('\n', ('', ), [(), ('', )])
+
+    def test_client_call_three_element_response(self):
+        # protocol.call() can get back tuples of other lengths. A three element
+        # tuple should be unpacked as three strings.
+        self.assertServerToClientEncoding('a\x01b\x0134\n', ('a', 'b', '34'),
+            [('a', 'b', '34')])
+
+    def test_client_call_with_body_bytes_uploads(self):
+        # protocol.call_with_upload should length-prefix the bytes onto the 
+        # wire.
+        expected_bytes = "foo\n7\nabcdefgdone\n"
+        input = StringIO("\n")
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call_with_body_bytes(('foo', ), "abcdefg")
+        self.assertEqual(expected_bytes, output.getvalue())
+
+    def test_client_call_with_body_readv_array(self):
+        # protocol.call_with_upload should encode the readv array and then
+        # length-prefix the bytes onto the wire.
+        expected_bytes = "foo\n7\n1,2\n5,6done\n"
+        input = StringIO("\n")
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call_with_body_readv_array(('foo', ), [(1,2),(5,6)])
+        self.assertEqual(expected_bytes, output.getvalue())
+
+    def test_client_read_body_bytes_all(self):
+        # read_body_bytes should decode the body bytes from the wire into
+        # a response.
+        expected_bytes = "1234567"
+        server_bytes = "ok\n7\n1234567done\n"
+        input = StringIO(server_bytes)
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call('foo')
+        protocol.read_response_tuple(True)
+        self.assertEqual(expected_bytes, protocol.read_body_bytes())
+
+    def test_client_read_body_bytes_incremental(self):
+        # test reading a few bytes at a time from the body
+        # XXX: possibly we should test dribbling the bytes into the stringio
+        # to make the state machine work harder: however, as we use the
+        # LengthPrefixedBodyDecoder that is already well tested - we can skip
+        # that.
+        expected_bytes = "1234567"
+        server_bytes = "ok\n7\n1234567done\n"
+        input = StringIO(server_bytes)
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call('foo')
+        protocol.read_response_tuple(True)
+        self.assertEqual(expected_bytes[0:2], protocol.read_body_bytes(2))
+        self.assertEqual(expected_bytes[2:4], protocol.read_body_bytes(2))
+        self.assertEqual(expected_bytes[4:6], protocol.read_body_bytes(2))
+        self.assertEqual(expected_bytes[6], protocol.read_body_bytes())
+
+    def test_client_cancel_read_body_does_not_eat_body_bytes(self):
+        # cancelling the expected body needs to finish the request, but not
+        # read any more bytes.
+        expected_bytes = "1234567"
+        server_bytes = "ok\n7\n1234567done\n"
+        input = StringIO(server_bytes)
+        output = StringIO()
+        medium = smart.SmartSimplePipesClientMedium(input, output)
+        protocol = smart.SmartClientRequestProtocolOne(medium.get_request())
+        protocol.call('foo')
+        protocol.read_response_tuple(True)
+        protocol.cancel_read_body()
+        self.assertEqual(3, input.tell())
+        self.assertRaises(errors.ReadingCompleted, protocol.read_body_bytes)
+
+
+class LengthPrefixedBodyDecoder(tests.TestCase):
+
+    # XXX: TODO: make accept_reading_trailer invoke translate_response or 
+    # something similar to the ProtocolBase method.
+
+    def test_construct(self):
+        decoder = smart.LengthPrefixedBodyDecoder()
+        self.assertFalse(decoder.finished_reading)
+        self.assertEqual(6, decoder.next_read_size())
+        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual('', decoder.unused_data)
+
+    def test_accept_bytes(self):
+        decoder = smart.LengthPrefixedBodyDecoder()
+        decoder.accept_bytes('')
+        self.assertFalse(decoder.finished_reading)
+        self.assertEqual(6, decoder.next_read_size())
+        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual('', decoder.unused_data)
+        decoder.accept_bytes('7')
+        self.assertFalse(decoder.finished_reading)
+        self.assertEqual(6, decoder.next_read_size())
+        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual('', decoder.unused_data)
+        decoder.accept_bytes('\na')
+        self.assertFalse(decoder.finished_reading)
+        self.assertEqual(11, decoder.next_read_size())
+        self.assertEqual('a', decoder.read_pending_data())
+        self.assertEqual('', decoder.unused_data)
+        decoder.accept_bytes('bcdefgd')
+        self.assertFalse(decoder.finished_reading)
+        self.assertEqual(4, decoder.next_read_size())
+        self.assertEqual('bcdefg', decoder.read_pending_data())
+        self.assertEqual('', decoder.unused_data)
+        decoder.accept_bytes('one')
+        self.assertFalse(decoder.finished_reading)
+        self.assertEqual(1, decoder.next_read_size())
+        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual('', decoder.unused_data)
+        decoder.accept_bytes('\nblarg')
+        self.assertTrue(decoder.finished_reading)
+        self.assertEqual(1, decoder.next_read_size())
+        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual('blarg', decoder.unused_data)
+        
+    def test_accept_bytes_all_at_once_with_excess(self):
+        decoder = smart.LengthPrefixedBodyDecoder()
+        decoder.accept_bytes('1\nadone\nunused')
+        self.assertTrue(decoder.finished_reading)
+        self.assertEqual(1, decoder.next_read_size())
+        self.assertEqual('a', decoder.read_pending_data())
+        self.assertEqual('unused', decoder.unused_data)
+
+    def test_accept_bytes_exact_end_of_body(self):
+        decoder = smart.LengthPrefixedBodyDecoder()
+        decoder.accept_bytes('1\na')
+        self.assertFalse(decoder.finished_reading)
+        self.assertEqual(5, decoder.next_read_size())
+        self.assertEqual('a', decoder.read_pending_data())
+        self.assertEqual('', decoder.unused_data)
+        decoder.accept_bytes('done\n')
+        self.assertTrue(decoder.finished_reading)
+        self.assertEqual(1, decoder.next_read_size())
+        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual('', decoder.unused_data)
+
+
+class FakeHTTPMedium(object):
+    def __init__(self):
+        self.written_request = None
+        self._current_request = None
+    def send_http_smart_request(self, bytes):
+        self.written_request = bytes
+        return None
+
+
+class HTTPTunnellingSmokeTest(tests.TestCaseWithTransport):
+    
+    def _test_bulk_data(self, url_protocol):
+        # We should be able to send and receive bulk data in a single message.
+        # The 'readv' command in the smart protocol both sends and receives bulk
+        # data, so we use that.
+        self.build_tree(['data-file'])
+        http_server = HTTPServerWithSmarts()
+        http_server._url_protocol = url_protocol
+        http_server.setUp()
+        self.addCleanup(http_server.tearDown)
+
+        http_transport = get_transport(http_server.get_url())
+
+        medium = http_transport.get_smart_medium()
+        #remote_transport = RemoteTransport('fake_url', medium)
+        remote_transport = smart.SmartTransport('/', medium=medium)
+        self.assertEqual(
+            [(0, "c")], list(remote_transport.readv("data-file", [(0,1)])))
+
+    def test_bulk_data_pycurl(self):
+        try:
+            self._test_bulk_data('http+pycurl')
+        except errors.UnsupportedProtocol, e:
+            raise tests.TestSkipped(str(e))
+    
+    def test_bulk_data_urllib(self):
+        self._test_bulk_data('http+urllib')
+
+    def test_smart_http_medium_request_accept_bytes(self):
+        medium = FakeHTTPMedium()
+        request = SmartClientHTTPMediumRequest(medium)
+        request.accept_bytes('abc')
+        request.accept_bytes('def')
+        self.assertEqual(None, medium.written_request)
+        request.finished_writing()
+        self.assertEqual('abcdef', medium.written_request)
+
+    def _test_http_send_smart_request(self, url_protocol):
+        http_server = HTTPServerWithSmarts()
+        http_server._url_protocol = url_protocol
+        http_server.setUp()
+        self.addCleanup(http_server.tearDown)
+
+        post_body = 'hello\n'
+        expected_reply_body = 'ok\x011\n'
+
+        http_transport = get_transport(http_server.get_url())
+        medium = http_transport.get_smart_medium()
+        response = medium.send_http_smart_request(post_body)
+        reply_body = response.read()
+        self.assertEqual(expected_reply_body, reply_body)
+
+    def test_http_send_smart_request_pycurl(self):
+        try:
+            self._test_http_send_smart_request('http+pycurl')
+        except errors.UnsupportedProtocol, e:
+            raise tests.TestSkipped(str(e))
+
+    def test_http_send_smart_request_urllib(self):
+        self._test_http_send_smart_request('http+urllib')
+
+    def test_http_server_with_smarts(self):
+        http_server = HTTPServerWithSmarts()
+        http_server.setUp()
+        self.addCleanup(http_server.tearDown)
+
+        post_body = 'hello\n'
+        expected_reply_body = 'ok\x011\n'
+
+        smart_server_url = http_server.get_url() + '.bzr/smart'
+        reply = urllib2.urlopen(smart_server_url, post_body).read()
+
+        self.assertEqual(expected_reply_body, reply)
+
+    def test_smart_http_server_post_request_handler(self):
+        http_server = HTTPServerWithSmarts()
+        http_server.setUp()
+        self.addCleanup(http_server.tearDown)
+        httpd = http_server._get_httpd()
+
+        socket = SampleSocket(
+            'POST /.bzr/smart HTTP/1.0\r\n'
+            # HTTP/1.0 posts must have a Content-Length.
+            'Content-Length: 6\r\n'
+            '\r\n'
+            'hello\n')
+        request_handler = SmartRequestHandler(
+            socket, ('localhost', 80), httpd)
+        response = socket.writefile.getvalue()
+        self.assertStartsWith(response, 'HTTP/1.0 200 ')
+        # This includes the end of the HTTP headers, and all the body.
+        expected_end_of_response = '\r\n\r\nok\x011\n'
+        self.assertEndsWith(response, expected_end_of_response)
+
+
+class SampleSocket(object):
+    """A socket-like object for use in testing the HTTP request handler."""
+    
+    def __init__(self, socket_read_content):
+        """Constructs a sample socket.
+
+        :param socket_read_content: a byte sequence
+        """
+        # Use plain python StringIO so we can monkey-patch the close method to
+        # not discard the contents.
+        from StringIO import StringIO
+        self.readfile = StringIO(socket_read_content)
+        self.writefile = StringIO()
+        self.writefile.close = lambda: None
+        
+    def makefile(self, mode='r', bufsize=None):
+        if 'r' in mode:
+            return self.readfile
+        else:
+            return self.writefile
+
+        
 # TODO: Client feature that does get_bundle and then installs that into a
 # branch; this can be used in place of the regular pull/fetch operation when
 # coming from a smart server.

=== modified file 'bzrlib/tests/test_transport_implementations.py'
--- bzrlib/tests/test_transport_implementations.py	2006-10-06 17:07:58 +0000
+++ bzrlib/tests/test_transport_implementations.py	2006-10-13 10:50:33 +0000
@@ -897,6 +897,8 @@
         except NotImplementedError:
             raise TestSkipped("Transport %s has no bogus URL support." %
                               self._server.__class__)
+        # This should be:  but SSH still connects on construction. No COOKIE!
+        # self.assertRaises((ConnectionError, NoSuchFile), t.get, '.bzr/branch')
         try:
             t = bzrlib.transport.get_transport(url)
             t.get('.bzr/branch')
@@ -1312,19 +1314,14 @@
         self.assertEqual(d[2], (0, '0'))
         self.assertEqual(d[3], (3, '34'))
 
-    def test_get_smart_client(self):
-        """All transports must either give a smart client, or know they can't.
-
-        For some transports such as http this might depend on probing to see 
-        what's actually present on the other end.  (But we can adjust for that 
-        in the future.)
+    def test_get_smart_medium(self):
+        """All transports must either give a smart medium, or know they can't.
         """
         transport = self.get_transport()
         try:
-            client = transport.get_smart_client()
-            # XXX: should be a more general class
-            self.assertIsInstance(client, smart.SmartStreamClient)
-        except NoSmartServer:
+            medium = transport.get_smart_medium()
+            self.assertIsInstance(medium, smart.SmartClientMedium)
+        except errors.NoSmartMedium:
             # as long as we got it we're fine
             pass
 

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2006-10-11 00:23:23 +0000
+++ bzrlib/transport/__init__.py	2006-10-13 06:41:04 +0000
@@ -446,10 +446,23 @@
     def get_smart_client(self):
         """Return a smart client for this transport if possible.
 
+        A smart client doesn't imply the presence of a smart server: it implies
+        that the smart protocol can be tunnelled via this transport.
+
         :raises NoSmartServer: if no smart server client is available.
         """
         raise errors.NoSmartServer(self.base)
 
+    def get_smart_medium(self):
+        """Return a smart client medium for this transport if possible.
+
+        A smart medium doesn't imply the presence of a smart server: it implies
+        that the smart protocol can be tunnelled via this transport.
+
+        :raises NoSmartMedium: if no smart server medium is available.
+        """
+        raise errors.NoSmartMedium(self)
+
     def readv(self, relpath, offsets):
         """Get parts of the file at the given relative path.
 

=== modified file 'bzrlib/transport/http/__init__.py'
--- bzrlib/transport/http/__init__.py	2006-10-12 14:29:32 +0000
+++ bzrlib/transport/http/__init__.py	2006-10-13 10:50:33 +0000
@@ -27,7 +27,10 @@
 
 from bzrlib import errors
 from bzrlib.trace import mutter
-from bzrlib.transport import Transport
+from bzrlib.transport import (
+    smart,
+    Transport,
+    )
 from bzrlib.ui import ui_factory
 
 
@@ -110,7 +113,7 @@
     return m
 
 
-class HttpTransportBase(Transport):
+class HttpTransportBase(Transport, smart.SmartClientMedium):
     """Base class for http implementations.
 
     Does URL parsing, etc, but not any network IO.
@@ -162,8 +165,9 @@
         else:
             # Except for the root, no trailing slashes are allowed
             if len(relpath_parts) > 1 and relpath_parts[-1] == '':
-                raise ValueError("path %r within branch %r seems to be a directory"
-                                 % (relpath, self._path))
+                raise ValueError(
+                    "path %r within branch %r seems to be a directory"
+                    % (relpath, self._path))
             basepath = self._path.split('/')
             if len(basepath) > 0 and basepath[-1] == '':
                 basepath = basepath[:-1]
@@ -226,6 +230,17 @@
         """
         raise NotImplementedError(self._get)
 
+    def get_request(self):
+        return SmartClientHTTPMediumRequest(self)
+
+    def get_smart_medium(self):
+        """See Transport.get_smart_medium.
+
+        HttpTransportBase directly implements the minimal interface of
+        SmartMediumClient, so this returns self.
+        """
+        return self
+
     def readv(self, relpath, offsets):
         """Get parts of the file at the given relative path.
 
@@ -272,6 +287,17 @@
 
         return combined
 
+    def _post(self, body_bytes):
+        """POST body_bytes to .bzr/smart on this transport.
+        
+        :returns: (response code, response body file-like object).
+        """
+        # TODO: Requiring all the body_bytes to be available at the beginning of
+        # the POST may require large client buffers.  It would be nice to have
+        # an interface that allows streaming via POST when possible (and
+        # degrades to a local buffer when not).
+        raise NotImplementedError(self._post)
+
     def put_file(self, relpath, f, mode=None):
         """Copy the file-like object into the location.
 
@@ -386,3 +412,30 @@
             strings.append('-%d' % tail_amount)
 
         return ','.join(strings)
+
+    def send_http_smart_request(self, bytes):
+        code, body_filelike = self._post(bytes)
+        assert code == 200, 'unexpected HTTP response code %r' % (code,)
+        return body_filelike
+
+
+class SmartClientHTTPMediumRequest(smart.SmartClientMediumRequest):
+    """A SmartClientMediumRequest that works with an HTTP medium."""
+
+    def __init__(self, medium):
+        smart.SmartClientMediumRequest.__init__(self, medium)
+        self._buffer = ''
+
+    def _accept_bytes(self, bytes):
+        self._buffer += bytes
+
+    def _finished_writing(self):
+        data = self._medium.send_http_smart_request(self._buffer)
+        self._response_body = data
+
+    def _read_bytes(self, count):
+        return self._response_body.read(count)
+
+    def _finished_reading(self):
+        """See SmartClientMediumRequest._finished_reading."""
+        pass

=== modified file 'bzrlib/transport/http/_pycurl.py'
--- bzrlib/transport/http/_pycurl.py	2006-10-13 08:53:34 +0000
+++ bzrlib/transport/http/_pycurl.py	2006-10-13 10:50:33 +0000
@@ -101,6 +101,7 @@
         abspath = self._real_abspath(relpath)
         curl.setopt(pycurl.URL, abspath)
         self._set_curl_options(curl)
+        curl.setopt(pycurl.HTTPGET, 1)
         # don't want the body - ie just do a HEAD request
         # This means "NO BODY" not 'nobody'
         curl.setopt(pycurl.NOBODY, 1)
@@ -126,6 +127,14 @@
             return self._get_full(relpath)
 
     def _setup_get_request(self, curl, relpath):
+        # Make sure we do a GET request. versions > 7.14.1 also set the
+        # NO BODY flag, but we'll do it ourselves in case it is an older
+        # pycurl version
+        curl.setopt(pycurl.NOBODY, 0)
+        curl.setopt(pycurl.HTTPGET, 1)
+        return self._setup_request(curl, relpath)
+
+    def _setup_request(self, curl, relpath):
         """Do the common setup stuff for making a request
 
         :param curl: The curl object to place the request on
@@ -138,11 +147,6 @@
         abspath = self._real_abspath(relpath)
         curl.setopt(pycurl.URL, abspath)
         self._set_curl_options(curl)
-        # Make sure we do a GET request. versions > 7.14.1 also set the
-        # NO BODY flag, but we'll do it ourselves in case it is an older
-        # pycurl version
-        curl.setopt(pycurl.NOBODY, 0)
-        curl.setopt(pycurl.HTTPGET, 1)
 
         data = StringIO()
         header = StringIO()
@@ -191,6 +195,22 @@
         # handle_response will raise NoSuchFile, etc based on the response code
         return code, response.handle_response(abspath, code, headers, data)
 
+    def _post(self, body_bytes):
+        fake_file = StringIO(body_bytes)
+        curl = self._base_curl
+        # Other places that use _base_curl for GET requests explicitly set
+        # HTTPGET, so it should be safe to re-use the same object for both GETs
+        # and POSTs.
+        curl.setopt(pycurl.POST, 1)
+        curl.setopt(pycurl.POSTFIELDSIZE, len(body_bytes))
+        curl.setopt(pycurl.READFUNCTION, fake_file.read)
+        abspath, data, header = self._setup_request(curl, '.bzr/smart')
+        self._curl_perform(curl)
+        data.seek(0)
+        code = curl.getinfo(pycurl.HTTP_CODE)
+        headers = _extract_headers(header.getvalue(), abspath)
+        return code, response.handle_response(abspath, code, headers, data)
+
     def _raise_curl_http_error(self, curl, info=None):
         code = curl.getinfo(pycurl.HTTP_CODE)
         url = curl.getinfo(pycurl.EFFECTIVE_URL)
@@ -211,9 +231,12 @@
         # There's no way in http/1.0 to say "must revalidate"; we don't want
         # to force it to always retrieve.  so just turn off the default Pragma
         # provided by Curl.
+        # Also, we override the Expect: header so that pycurl will send the POST
+        # body immediately.
         headers = ['Cache-control: max-age=0',
                    'Pragma: no-cache',
-                   'Connection: Keep-Alive']
+                   'Connection: Keep-Alive',
+                   'Expect: ',]
         ## curl.setopt(pycurl.VERBOSE, 1)
         # TODO: maybe include a summary of the pycurl version
         ua_str = 'bzr/%s (pycurl)' % (bzrlib.__version__,)

=== modified file 'bzrlib/transport/http/_urllib.py'
--- bzrlib/transport/http/_urllib.py	2006-10-12 14:29:32 +0000
+++ bzrlib/transport/http/_urllib.py	2006-10-13 10:50:33 +0000
@@ -81,7 +81,10 @@
                                               request.user, request.password)
 
     def _perform(self, request):
-        """Send the request to the server and handles common errors."""
+        """Send the request to the server and handles common errors.
+
+        :returns: urllib2 Response object
+        """
         if self._connection is not None:
             # Give back shared info
             request.connection = self._connection
@@ -137,6 +140,15 @@
         self._connection.fake_close()
         return code, data
 
+    def _post(self, body_bytes):
+        abspath = self._real_abspath('.bzr/smart')
+        response = self._perform(Request('POST', abspath, body_bytes))
+        code = response.code
+        data = handle_response(abspath, code, response.headers, response)
+        # Close response to free the httplib.HTTPConnection pipeline
+        self._connection.fake_close()
+        return code, data
+
     def should_cache(self):
         """Return True if the data pulled across should be cached locally.
         """

=== modified file 'bzrlib/transport/http/response.py'
--- bzrlib/transport/http/response.py	2006-10-11 08:51:52 +0000
+++ bzrlib/transport/http/response.py	2006-10-13 06:36:44 +0000
@@ -251,7 +251,11 @@
         try:
             content_type = headers['Content-Type']
         except KeyError:
-            # We can't be multipart if there is no Content-Type header
+            # When there is no content-type header we treat
+            # the response as being of type 'application/octet-stream' as per
+            # RFC2616 section 7.2.1.
+            # Therefore it is obviously not multipart
+            content_type = 'application/octet-stream'
             is_multipart = False
         else:
             is_multipart = _is_multipart(content_type)

=== modified file 'bzrlib/transport/smart.py'
--- bzrlib/transport/smart.py	2006-10-04 07:57:23 +0000
+++ bzrlib/transport/smart.py	2006-10-09 07:34:15 +0000
@@ -22,7 +22,7 @@
 
   SEP := '\001'
     Fields are separated by Ctrl-A.
-  BULK_DATA := CHUNK+ TRAILER
+  BULK_DATA := CHUNK TRAILER
     Chunks can be repeated as many times as necessary.
   CHUNK := CHUNK_LEN CHUNK_BODY
   CHUNK_LEN := DIGIT+ NEWLINE
@@ -46,6 +46,66 @@
 URLs that include ~ should probably be passed across to the server verbatim
 and the server can expand them.  This will proably not be meaningful when 
 limited to a directory?
+
+At the bottom level socket, pipes, HTTP server.  For sockets, we have the idea
+that you have multiple requests and get a read error because the other side did
+shutdown.  For pipes we have read pipe which will have a zero read which marks
+end-of-file.  For HTTP server environment there is not end-of-stream because
+each request coming into the server is independent.
+
+So we need a wrapper around pipes and sockets to seperate out requests from
+substrate and this will give us a single model which is consist for HTTP,
+sockets and pipes.
+
+Server-side
+-----------
+
+ MEDIUM  (factory for protocol, reads bytes & pushes to protocol,
+          uses protocol to detect end-of-request, sends written
+          bytes to client) e.g. socket, pipe, HTTP request handler.
+  ^
+  | bytes.
+  v
+
+PROTOCOL  (serialization, deserialization)  accepts bytes for one
+          request, decodes according to internal state, pushes
+          structured data to handler.  accepts structured data from
+          handler and encodes and writes to the medium.  factory for
+          handler.
+  ^
+  | structured data
+  v
+
+HANDLER   (domain logic) accepts structured data, operates state
+          machine until the request can be satisfied,
+          sends structured data to the protocol.
+
+
+Client-side
+-----------
+
+ CLIENT             domain logic, accepts domain requests, generated structured
+                    data, reads structured data from responses and turns into
+                    domain data.  Sends structured data to the protocol.
+                    Operates state machines until the request can be delivered
+                    (e.g. reading from a bundle generated in bzrlib to deliver a
+                    complete request).
+
+                    Possibly this should just be RemoteBzrDir, RemoteTransport,
+                    ...
+  ^
+  | structured data
+  v
+
+PROTOCOL  (serialization, deserialization)  accepts structured data for one
+          request, encodes and writes to the medium.  Reads bytes from the
+          medium, decodes and allows the client to read structured data.
+  ^
+  | bytes.
+  v
+
+ MEDIUM  (accepts bytes from the protocol & delivers to the remote server.
+          Allows the potocol to read bytes e.g. socket, pipe, HTTP request.
 """
 
 
@@ -133,18 +193,10 @@
 # TODO: SmartBzrDir class, proxying all Branch etc methods across to another
 # branch doing file-level operations.
 #
-# TODO: jam 20060915 _decode_tuple is acting directly on input over
-#       the socket, and it assumes everything is UTF8 sections separated
-#       by \001. Which means a request like '\002' Will abort the connection
-#       because of a UnicodeDecodeError. It does look like invalid data will
-#       kill the SmartStreamServer, but only with an abort + exception, and 
-#       the overall server shouldn't die.
 
 from cStringIO import StringIO
-import errno
 import os
 import socket
-import sys
 import tempfile
 import threading
 import urllib
@@ -159,8 +211,11 @@
     urlutils,
     )
 from bzrlib.bundle.serializer import write_bundle
-from bzrlib.trace import mutter
-from bzrlib.transport import local
+try:
+    from bzrlib.transport import ssh
+except errors.ParamikoNotPresent:
+    # no paramiko.  SmartSSHClientMedium will break.
+    pass
 
 # must do this otherwise urllib can't parse the urls properly :(
 for scheme in ['ssh', 'bzr', 'bzr+loopback', 'bzr+ssh']:
@@ -178,7 +233,12 @@
         return None
     if req_line[-1] != '\n':
         raise errors.SmartProtocolError("request %r not terminated" % req_line)
-    return tuple((a.decode('utf-8') for a in req_line[:-1].split('\x01')))
+    try:
+        return tuple((a.decode('utf-8') for a in req_line[:-1].split('\x01')))
+    except UnicodeDecodeError:
+        raise errors.SmartProtocolError(
+            "one or more arguments of request %r are not valid UTF-8"
+            % req_line)
 
 
 def _encode_tuple(args):
@@ -189,35 +249,11 @@
 class SmartProtocolBase(object):
     """Methods common to client and server"""
 
-    def _send_bulk_data(self, body):
-        """Send chunked body data"""
-        assert isinstance(body, str)
-        bytes = ''.join(('%d\n' % len(body), body, 'done\n'))
-        self._write_and_flush(bytes)
-
-    # TODO: this only actually accomodates a single block; possibly should support
-    # multiple chunks?
-    def _recv_bulk(self):
-        chunk_len = self._in.readline()
-        try:
-            chunk_len = int(chunk_len)
-        except ValueError:
-            raise errors.SmartProtocolError("bad chunk length line %r" % chunk_len)
-        bulk = self._in.read(chunk_len)
-        if len(bulk) != chunk_len:
-            raise errors.SmartProtocolError("short read fetching bulk data chunk")
-        self._recv_trailer()
-        return bulk
-
-    def _recv_tuple(self):
-        return _recv_tuple(self._in)
-
-    def _recv_trailer(self):
-        resp = self._recv_tuple()
-        if resp == ('done', ):
-            return
-        else:
-            self._translate_error(resp)
+    # TODO: this only actually accomodates a single block; possibly should
+    # support multiple chunks?
+    def _encode_bulk_data(self, body):
+        """Encode body as a bulk data chunk."""
+        return ''.join(('%d\n' % len(body), body, 'done\n'))
 
     def _serialise_offsets(self, offsets):
         """Serialise a readv offset list."""
@@ -225,15 +261,193 @@
         for start, length in offsets:
             txt.append('%d,%d' % (start, length))
         return '\n'.join(txt)
-
-    def _write_and_flush(self, bytes):
-        """Write bytes to self._out and flush it."""
-        # XXX: this will be inefficient.  Just ask Robert.
-        self._out.write(bytes)
-        self._out.flush()
-
-
-class SmartStreamServer(SmartProtocolBase):
+        
+
+class SmartServerRequestProtocolOne(SmartProtocolBase):
+    """Server-side encoding and decoding logic for smart version 1."""
+    
+    def __init__(self, backing_transport, write_func):
+        self._backing_transport = backing_transport
+        self.excess_buffer = ''
+        self._finished_reading = False
+        self.in_buffer = ''
+        self.has_dispatched = False
+        self.request = None
+        self._body_decoder = None
+        self._write_func = write_func
+
+    def accept_bytes(self, bytes):
+        """Take bytes, and advance the internal state machine appropriately.
+        
+        :param bytes: must be a byte string
+        """
+        assert isinstance(bytes, str)
+        self.in_buffer += bytes
+        if not self.has_dispatched:
+            if '\n' not in self.in_buffer:
+                # no command line yet
+                return
+            self.has_dispatched = True
+            try:
+                first_line, self.in_buffer = self.in_buffer.split('\n', 1)
+                first_line += '\n'
+                req_args = _decode_tuple(first_line)
+                self.request = SmartServerRequestHandler(
+                    self._backing_transport)
+                self.request.dispatch_command(req_args[0], req_args[1:])
+                if self.request.finished_reading:
+                    # trivial request
+                    self.excess_buffer = self.in_buffer
+                    self.in_buffer = ''
+                    self._send_response(self.request.response.args,
+                        self.request.response.body)
+                self.sync_with_request(self.request)
+            except KeyboardInterrupt:
+                raise
+            except Exception, exception:
+                # everything else: pass to client, flush, and quit
+                self._send_response(('error', str(exception)))
+                return None
+
+        if self.has_dispatched:
+            if self._finished_reading:
+                # nothing to do.XXX: this routine should be a single state 
+                # machine too.
+                self.excess_buffer += self.in_buffer
+                self.in_buffer = ''
+                return
+            if self._body_decoder is None:
+                self._body_decoder = LengthPrefixedBodyDecoder()
+            self._body_decoder.accept_bytes(self.in_buffer)
+            self.in_buffer = self._body_decoder.unused_data
+            body_data = self._body_decoder.read_pending_data()
+            self.request.accept_body(body_data)
+            if self._body_decoder.finished_reading:
+                self.request.end_of_body()
+                assert self.request.finished_reading, \
+                    "no more body, request not finished"
+            self.sync_with_request(self.request)
+            if self.request.response is not None:
+                self._send_response(self.request.response.args,
+                    self.request.response.body)
+                self.excess_buffer = self.in_buffer
+                self.in_buffer = ''
+            else:
+                assert not self.request.finished_reading, \
+                    "no response and we have finished reading."
+
+    def _send_response(self, args, body=None):
+        """Send a smart server response down the output stream."""
+        self._write_func(_encode_tuple(args))
+        if body is not None:
+            assert isinstance(body, str), 'body must be a str'
+            bytes = self._encode_bulk_data(body)
+            self._write_func(bytes)
+
+    def sync_with_request(self, request):
+        self._finished_reading = request.finished_reading
+        
+    def next_read_size(self):
+        if self._finished_reading:
+            return 0
+        if self._body_decoder is None:
+            return 1
+        else:
+            return self._body_decoder.next_read_size()
+
+
+class LengthPrefixedBodyDecoder(object):
+    """Decodes the length-prefixed bulk data."""
+    
+    def __init__(self):
+        self.bytes_left = None
+        self.finished_reading = False
+        self.unused_data = ''
+        self.state_accept = self._state_accept_expecting_length
+        self.state_read = self._state_read_no_data
+        self._in_buffer = ''
+        self._trailer_buffer = ''
+    
+    def accept_bytes(self, bytes):
+        """Decode as much of bytes as possible.
+
+        If 'bytes' contains too much data it will be appended to
+        self.unused_data.
+
+        finished_reading will be set when no more data is required.  Further
+        data will be appended to self.unused_data.
+        """
+        # accept_bytes is allowed to change the state
+        current_state = self.state_accept
+        self.state_accept(bytes)
+        while current_state != self.state_accept:
+            current_state = self.state_accept
+            self.state_accept('')
+
+    def next_read_size(self):
+        if self.bytes_left is not None:
+            # Ideally we want to read all the remainder of the body and the
+            # trailer in one go.
+            return self.bytes_left + 5
+        elif self.state_accept == self._state_accept_reading_trailer:
+            # Just the trailer left
+            return 5 - len(self._trailer_buffer)
+        elif self.state_accept == self._state_accept_expecting_length:
+            # There's still at least 6 bytes left ('\n' to end the length, plus
+            # 'done\n').
+            return 6
+        else:
+            # Reading excess data.  Either way, 1 byte at a time is fine.
+            return 1
+        
+    def read_pending_data(self):
+        """Return any pending data that has been decoded."""
+        return self.state_read()
+
+    def _state_accept_expecting_length(self, bytes):
+        self._in_buffer += bytes
+        pos = self._in_buffer.find('\n')
+        if pos == -1:
+            return
+        self.bytes_left = int(self._in_buffer[:pos])
+        self._in_buffer = self._in_buffer[pos+1:]
+        self.bytes_left -= len(self._in_buffer)
+        self.state_accept = self._state_accept_reading_body
+        self.state_read = self._state_read_in_buffer
+
+    def _state_accept_reading_body(self, bytes):
+        self._in_buffer += bytes
+        self.bytes_left -= len(bytes)
+        if self.bytes_left <= 0:
+            # Finished with body
+            if self.bytes_left != 0:
+                self._trailer_buffer = self._in_buffer[self.bytes_left:]
+                self._in_buffer = self._in_buffer[:self.bytes_left]
+            self.bytes_left = None
+            self.state_accept = self._state_accept_reading_trailer
+        
+    def _state_accept_reading_trailer(self, bytes):
+        self._trailer_buffer += bytes
+        # TODO: what if the trailer does not match "done\n"?  Should this raise
+        # a ProtocolViolation exception?
+        if self._trailer_buffer.startswith('done\n'):
+            self.unused_data = self._trailer_buffer[len('done\n'):]
+            self.state_accept = self._state_accept_reading_unused
+            self.finished_reading = True
+    
+    def _state_accept_reading_unused(self, bytes):
+        self.unused_data += bytes
+
+    def _state_read_no_data(self):
+        return ''
+
+    def _state_read_in_buffer(self):
+        result = self._in_buffer
+        self._in_buffer = ''
+        return result
+
+
+class SmartServerStreamMedium(object):
     """Handles smart commands coming over a stream.
 
     The stream may be a pipe connected to sshd, or a tcp socket, or an
@@ -246,6 +460,86 @@
     which will typically be a LocalTransport looking at the server's filesystem.
     """
 
+    def __init__(self, backing_transport):
+        """Construct new server.
+
+        :param backing_transport: Transport for the directory served.
+        """
+        # backing_transport could be passed to serve instead of __init__
+        self.backing_transport = backing_transport
+        self.finished = False
+
+    def serve(self):
+        """Serve requests until the client disconnects."""
+        # Keep a reference to stderr because the sys module's globals get set to
+        # None during interpreter shutdown.
+        from sys import stderr
+        try:
+            while not self.finished:
+                protocol = SmartServerRequestProtocolOne(self.backing_transport,
+                                                         self._write_out)
+                self._serve_one_request(protocol)
+        except Exception, e:
+            stderr.write("%s terminating on exception %s\n" % (self, e))
+            raise
+
+    def _serve_one_request(self, protocol):
+        """Read one request from input, process, send back a response.
+        
+        :param protocol: a SmartServerRequestProtocol.
+        """
+        try:
+            self._serve_one_request_unguarded(protocol)
+        except KeyboardInterrupt:
+            raise
+        except Exception, e:
+            self.terminate_due_to_error()
+
+    def terminate_due_to_error(self):
+        """Called when an unhandled exception from the protocol occurs."""
+        raise NotImplementedError(self.terminate_due_to_error)
+
+
+class SmartServerSocketStreamMedium(SmartServerStreamMedium):
+
+    def __init__(self, sock, backing_transport):
+        """Constructor.
+
+        :param sock: the socket the server will read from.  It will be put
+            into blocking mode.
+        """
+        SmartServerStreamMedium.__init__(self, backing_transport)
+        self.push_back = ''
+        sock.setblocking(True)
+        self.socket = sock
+
+    def _serve_one_request_unguarded(self, protocol):
+        while protocol.next_read_size():
+            if self.push_back:
+                protocol.accept_bytes(self.push_back)
+                self.push_back = ''
+            else:
+                bytes = self.socket.recv(4096)
+                if bytes == '':
+                    self.finished = True
+                    return
+                protocol.accept_bytes(bytes)
+        
+        self.push_back = protocol.excess_buffer
+    
+    def terminate_due_to_error(self):
+        """Called when an unhandled exception from the protocol occurs."""
+        # TODO: This should log to a server log file, but no such thing
+        # exists yet.  Andrew Bennetts 2006-09-29.
+        self.socket.close()
+        self.finished = True
+
+    def _write_out(self, bytes):
+        self.socket.sendall(bytes)
+
+
+class SmartServerPipeStreamMedium(SmartServerStreamMedium):
+
     def __init__(self, in_file, out_file, backing_transport):
         """Construct new server.
 
@@ -253,95 +547,88 @@
         :param out_file: Python file to write responses.
         :param backing_transport: Transport for the directory served.
         """
+        SmartServerStreamMedium.__init__(self, backing_transport)
         self._in = in_file
         self._out = out_file
-        self.smart_server = SmartServer(backing_transport)
-        # server can call back to us to get bulk data - this is not really
-        # ideal, they should get it per request instead
-        self.smart_server._recv_body = self._recv_bulk
-
-    def _recv_tuple(self):
-        """Read a request from the client and return as a tuple.
-        
-        Returns None at end of file (if the client closed the connection.)
-        """
-        return _recv_tuple(self._in)
-
-    def _send_tuple(self, args):
-        """Send response header"""
-        return self._write_and_flush(_encode_tuple(args))
-
-    def _send_error_and_disconnect(self, exception):
-        self._send_tuple(('error', str(exception)))
-        ## self._out.close()
-        ## self._in.close()
-
-    def _serve_one_request(self):
-        """Read one request from input, process, send back a response.
-        
-        :return: False if the server should terminate, otherwise None.
-        """
-        req_args = self._recv_tuple()
-        if req_args == None:
-            # client closed connection
-            return False  # shutdown server
-        try:
-            response = self.smart_server.dispatch_command(req_args[0], req_args[1:])
-            self._send_tuple(response.args)
-            if response.body is not None:
-                self._send_bulk_data(response.body)
-        except KeyboardInterrupt:
-            raise
-        except Exception, e:
-            # everything else: pass to client, flush, and quit
-            self._send_error_and_disconnect(e)
-            return False
-
-    def serve(self):
-        """Serve requests until the client disconnects."""
-        # Keep a reference to stderr because the sys module's globals get set to
-        # None during interpreter shutdown.
-        from sys import stderr
-        try:
-            while self._serve_one_request() != False:
-                pass
-        except Exception, e:
-            stderr.write("%s terminating on exception %s\n" % (self, e))
-            raise
+
+    def _serve_one_request_unguarded(self, protocol):
+        while True:
+            bytes_to_read = protocol.next_read_size()
+            if bytes_to_read == 0:
+                # Finished serving this request.
+                self._out.flush()
+                return
+            bytes = self._in.read(bytes_to_read)
+            if bytes == '':
+                # Connection has been closed.
+                self.finished = True
+                self._out.flush()
+                return
+            protocol.accept_bytes(bytes)
+
+    def terminate_due_to_error(self):
+        # TODO: This should log to a server log file, but no such thing
+        # exists yet.  Andrew Bennetts 2006-09-29.
+        self._out.close()
+        self.finished = True
+
+    def _write_out(self, bytes):
+        self._out.write(bytes)
 
 
 class SmartServerResponse(object):
-    """Response generated by SmartServer."""
+    """Response generated by SmartServerRequestHandler."""
 
     def __init__(self, args, body=None):
         self.args = args
         self.body = body
 
-# XXX: TODO: Create a SmartServerRequest which will take the responsibility
+# XXX: TODO: Create a SmartServerRequestHandler which will take the responsibility
 # for delivering the data for a request. This could be done with as the
 # StreamServer, though that would create conflation between request and response
 # which may be undesirable.
 
 
-class SmartServer(object):
+class SmartServerRequestHandler(object):
     """Protocol logic for smart server.
     
     This doesn't handle serialization at all, it just processes requests and
     creates responses.
     """
 
-    # IMPORTANT FOR IMPLEMENTORS: It is important that SmartServer not contain
-    # encoding or decoding logic to allow the wire protocol to vary from the
-    # object protocol: we will want to tweak the wire protocol separate from
-    # the object model, and ideally we will be able to do that without having
-    # a SmartServer subclass for each wire protocol, rather just a Protocol
-    # subclass.
+    # IMPORTANT FOR IMPLEMENTORS: It is important that SmartServerRequestHandler
+    # not contain encoding or decoding logic to allow the wire protocol to vary
+    # from the object protocol: we will want to tweak the wire protocol separate
+    # from the object model, and ideally we will be able to do that without
+    # having a SmartServerRequestHandler subclass for each wire protocol, rather
+    # just a Protocol subclass.
 
     # TODO: Better way of representing the body for commands that take it,
     # and allow it to be streamed into the server.
     
     def __init__(self, backing_transport):
         self._backing_transport = backing_transport
+        self._converted_command = False
+        self.finished_reading = False
+        self._body_bytes = ''
+        self.response = None
+
+    def accept_body(self, bytes):
+        """Accept body data.
+
+        This should be overriden for each command that desired body data to
+        handle the right format of that data. I.e. plain bytes, a bundle etc.
+
+        The deserialisation into that format should be done in the Protocol
+        object. Set self.desired_body_format to the format your method will
+        handle.
+        """
+        # default fallback is to accumulate bytes.
+        self._body_bytes += bytes
+        
+    def _end_of_body_handler(self):
+        """An unimplemented end of body handler."""
+        raise NotImplementedError(self._end_of_body_handler)
         
     def do_hello(self):
         """Answer a version request with my version."""
@@ -363,9 +650,15 @@
             return int(mode)
 
     def do_append(self, relpath, mode):
+        self._converted_command = True
+        self._relpath = relpath
+        self._mode = self._deserialise_optional_mode(mode)
+        self._end_of_body_handler = self._handle_do_append_end
+    
+    def _handle_do_append_end(self):
         old_length = self._backing_transport.append_bytes(
-            relpath, self._recv_body(), self._deserialise_optional_mode(mode))
-        return SmartServerResponse(('appended', '%d' % old_length))
+            self._relpath, self._body_bytes, self._mode)
+        self.response = SmartServerResponse(('appended', '%d' % old_length))
 
     def do_delete(self, relpath):
         self._backing_transport.delete(relpath)
@@ -389,9 +682,15 @@
         self._backing_transport.move(rel_from, rel_to)
 
     def do_put(self, relpath, mode):
-        self._backing_transport.put_bytes(relpath,
-                self._recv_body(),
-                self._deserialise_optional_mode(mode))
+        self._converted_command = True
+        self._relpath = relpath
+        self._mode = self._deserialise_optional_mode(mode)
+        self._end_of_body_handler = self._handle_do_put
+
+    def _handle_do_put(self):
+        self._backing_transport.put_bytes(self._relpath,
+                self._body_bytes, self._mode)
+        self.response = SmartServerResponse(('ok',))
 
     def _deserialise_offsets(self, text):
         # XXX: FIXME this should be on the protocol object.
@@ -404,18 +703,39 @@
         return offsets
 
     def do_put_non_atomic(self, relpath, mode, create_parent, dir_mode):
-        create_parent_dir = (create_parent == 'T')
-        self._backing_transport.put_bytes_non_atomic(relpath,
-                self._recv_body(),
-                mode=self._deserialise_optional_mode(mode),
-                create_parent_dir=create_parent_dir,
-                dir_mode=self._deserialise_optional_mode(dir_mode))
+        self._converted_command = True
+        self._end_of_body_handler = self._handle_put_non_atomic
+        self._relpath = relpath
+        self._dir_mode = self._deserialise_optional_mode(dir_mode)
+        self._mode = self._deserialise_optional_mode(mode)
+        # a boolean would be nicer XXX
+        self._create_parent = (create_parent == 'T')
+
+    def _handle_put_non_atomic(self):
+        self._backing_transport.put_bytes_non_atomic(self._relpath,
+                self._body_bytes,
+                mode=self._mode,
+                create_parent_dir=self._create_parent,
+                dir_mode=self._dir_mode)
+        self.response = SmartServerResponse(('ok',))
 
     def do_readv(self, relpath):
-        offsets = self._deserialise_offsets(self._recv_body())
+        self._converted_command = True
+        self._end_of_body_handler = self._handle_readv_offsets
+        self._relpath = relpath
+
+    def end_of_body(self):
+        """No more body data will be received."""
+        self._run_handler_code(self._end_of_body_handler, (), {})
+        # cannot read after this.
+        self.finished_reading = True
+
+    def _handle_readv_offsets(self):
+        """accept offsets for a readv request."""
+        offsets = self._deserialise_offsets(self._body_bytes)
         backing_bytes = ''.join(bytes for offset, bytes in
-                             self._backing_transport.readv(relpath, offsets))
-        return SmartServerResponse(('readv',), backing_bytes)
+            self._backing_transport.readv(self._relpath, offsets))
+        self.response = SmartServerResponse(('readv',), backing_bytes)
         
     def do_rename(self, rel_from, rel_to):
         self._backing_transport.rename(rel_from, rel_to)
@@ -439,14 +759,35 @@
         return SmartServerResponse((), tmpf.read())
 
     def dispatch_command(self, cmd, args):
+        """Deprecated compatibility method.""" # XXX XXX
         func = getattr(self, 'do_' + cmd, None)
         if func is None:
             raise errors.SmartProtocolError("bad request %r" % (cmd,))
+        self._run_handler_code(func, args, {})
+
+    def _run_handler_code(self, callable, args, kwargs):
+        """Run some handler specific code 'callable'.
+
+        If a result is returned, it is considered to be the commands response,
+        and finished_reading is set true, and its assigned to self.response.
+
+        Any exceptions caught are translated and a response object created
+        from them.
+        """
+        result = self._call_converting_errors(callable, args, kwargs)
+        if result is not None:
+            self.response = result
+            self.finished_reading = True
+        # handle unconverted commands
+        if not self._converted_command:
+            self.finished_reading = True
+            if result is None:
+                self.response = SmartServerResponse(('ok',))
+
+    def _call_converting_errors(self, callable, args, kwargs):
+        """Call callable converting errors to Response objects."""
         try:
-            result = func(*args)
-            if result is None: 
-                result = SmartServerResponse(('ok',))
-            return result
+            return callable(*args, **kwargs)
         except errors.NoSuchFile, e:
             return SmartServerResponse(('NoSuchFile', e.path))
         except errors.FileExists, e:
@@ -477,7 +818,7 @@
 class SmartTCPServer(object):
     """Listens on a TCP socket and accepts connections from smart clients"""
 
-    def __init__(self, backing_transport=None, host='127.0.0.1', port=0):
+    def __init__(self, backing_transport, host='127.0.0.1', port=0):
         """Construct a new server.
 
         To actually start it running, call either start_background_thread or
@@ -486,8 +827,6 @@
         :param host: Name of the interface to listen on.
         :param port: TCP port to listen on, or 0 to allocate a transient port.
         """
-        if backing_transport is None:
-            backing_transport = memory.MemoryTransport()
         self._server_socket = socket.socket()
         self._server_socket.bind((host, port))
         self.port = self._server_socket.getsockname()[1]
@@ -522,10 +861,7 @@
         # propogates to the newly accepted socket.
         conn.setblocking(True)
         conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
-        from_client = conn.makefile('r')
-        to_client = conn.makefile('w')
-        handler = SmartStreamServer(from_client, to_client,
-                self.backing_transport)
+        handler = SmartServerSocketStreamMedium(conn, self.backing_transport)
         connection_thread = threading.Thread(None, handler.serve, name='smart-server-child')
         connection_thread.setDaemon(True)
         connection_thread.start()
@@ -608,10 +944,11 @@
     # SmartTransport is an adapter from the Transport object model to the 
     # SmartClient model, not an encoder.
 
-    def __init__(self, url, clone_from=None, client=None):
+    def __init__(self, url, clone_from=None, medium=None):
         """Constructor.
 
-        :param client: ignored when clone_from is not None.
+        :param medium: The medium to use for this RemoteTransport. This must be
+            supplied if clone_from is None.
         """
         ### Technically super() here is faulty because Transport's __init__
         ### fails to take 2 parameters, and if super were to choose a silly
@@ -622,16 +959,14 @@
         self._scheme, self._username, self._password, self._host, self._port, self._path = \
                 transport.split_url(url)
         if clone_from is None:
-            if client is None:
-                self._client = SmartStreamClient(self._connect_to_server)
-            else:
-                self._client = client
+            self._medium = medium
         else:
             # credentials may be stripped from the base in some circumstances
             # as yet to be clearly defined or documented, so copy them.
             self._username = clone_from._username
             # reuse same connection
-            self._client = clone_from._client
+            self._medium = clone_from._medium
+        assert self._medium is not None
 
     def abspath(self, relpath):
         """Return the full url to the given relative path.
@@ -647,16 +982,19 @@
         This essentially opens a handle on a different remote directory.
         """
         if relative_url is None:
-            return self.__class__(self.base, self)
+            return SmartTransport(self.base, self)
         else:
-            return self.__class__(self.abspath(relative_url), self)
+            return SmartTransport(self.abspath(relative_url), self)
 
     def is_readonly(self):
         """Smart server transport can do read/write file operations."""
         return False
                                                    
     def get_smart_client(self):
-        return self._client
+        return self._medium
+
+    def get_smart_medium(self):
+        return self._medium
                                                    
     def _unparse_url(self, path):
         """Return URL for a path.
@@ -679,12 +1017,28 @@
         """Returns the Unicode version of the absolute path for relpath."""
         return self._combine_paths(self._path, relpath)
 
+    def _call(self, method, *args):
+        resp = self._call2(method, *args)
+        self._translate_error(resp)
+
+    def _call2(self, method, *args):
+        """Call a method on the remote server."""
+        protocol = SmartClientRequestProtocolOne(self._medium.get_request())
+        protocol.call(method, *args)
+        return protocol.read_response_tuple()
+
+    def _call_with_body_bytes(self, method, args, body):
+        """Call a method on the remote server with body bytes."""
+        protocol = SmartClientRequestProtocolOne(self._medium.get_request())
+        protocol.call_with_body_bytes((method, ) + args, body)
+        return protocol.read_response_tuple()
+
     def has(self, relpath):
         """Indicate whether a remote file of the given name exists or not.
 
         :see: Transport.has()
         """
-        resp = self._client._call('has', self._remote_path(relpath))
+        resp = self._call2('has', self._remote_path(relpath))
         if resp == ('yes', ):
             return True
         elif resp == ('no', ):
@@ -697,11 +1051,17 @@
         
         :see: Transport.get_bytes()/get_file()
         """
+        return StringIO(self.get_bytes(relpath))
+
+    def get_bytes(self, relpath):
         remote = self._remote_path(relpath)
-        resp = self._client._call('get', remote)
+        protocol = SmartClientRequestProtocolOne(self._medium.get_request())
+        protocol.call('get', remote)
+        resp = protocol.read_response_tuple(True)
         if resp != ('ok', ):
+            protocol.cancel_read_body()
             self._translate_error(resp, relpath)
-        return StringIO(self._client._recv_bulk())
+        return protocol.read_body_bytes()
 
     def _serialise_optional_mode(self, mode):
         if mode is None:
@@ -710,17 +1070,15 @@
             return '%d' % mode
 
     def mkdir(self, relpath, mode=None):
-        resp = self._client._call('mkdir', 
-                                  self._remote_path(relpath), 
-                                  self._serialise_optional_mode(mode))
+        resp = self._call2('mkdir', self._remote_path(relpath),
+            self._serialise_optional_mode(mode))
         self._translate_error(resp)
 
     def put_bytes(self, relpath, upload_contents, mode=None):
         # FIXME: upload_file is probably not safe for non-ascii characters -
         # should probably just pass all parameters as length-delimited
         # strings?
-        resp = self._client._call_with_upload(
-            'put',
+        resp = self._call_with_body_bytes('put',
             (self._remote_path(relpath), self._serialise_optional_mode(mode)),
             upload_contents)
         self._translate_error(resp)
@@ -734,7 +1092,7 @@
         if create_parent_dir:
             create_parent_str = 'T'
 
-        resp = self._client._call_with_upload(
+        resp = self._call_with_body_bytes(
             'put_non_atomic',
             (self._remote_path(relpath), self._serialise_optional_mode(mode),
              create_parent_str, self._serialise_optional_mode(dir_mode)),
@@ -763,7 +1121,7 @@
         return self.append_bytes(relpath, from_file.read(), mode)
         
     def append_bytes(self, relpath, bytes, mode=None):
-        resp = self._client._call_with_upload(
+        resp = self._call_with_body_bytes(
             'append',
             (self._remote_path(relpath), self._serialise_optional_mode(mode)),
             bytes)
@@ -772,7 +1130,7 @@
         self._translate_error(resp)
 
     def delete(self, relpath):
-        resp = self._client._call('delete', self._remote_path(relpath))
+        resp = self._call2('delete', self._remote_path(relpath))
         self._translate_error(resp)
 
     def readv(self, relpath, offsets):
@@ -789,18 +1147,20 @@
                                limit=self._max_readv_combine,
                                fudge_factor=self._bytes_to_read_before_seek))
 
-
-        resp = self._client._call_with_upload(
-            'readv',
-            (self._remote_path(relpath),),
-            self._client._serialise_offsets((c.start, c.length) for c in coalesced))
+        protocol = SmartClientRequestProtocolOne(self._medium.get_request())
+        protocol.call_with_body_readv_array(
+            ('readv', self._remote_path(relpath)),
+            [(c.start, c.length) for c in coalesced])
+        resp = protocol.read_response_tuple(True)
 
         if resp[0] != 'readv':
             # This should raise an exception
+            protocol.cancel_read_body()
             self._translate_error(resp)
             return
 
-        data = self._client._recv_bulk()
+        # FIXME: this should know how many bytes are needed, for clarity.
+        data = protocol.read_body_bytes()
         # Cache the results, but only until they have been fulfilled
         data_map = {}
         for c_offset in coalesced:
@@ -819,22 +1179,18 @@
                 cur_offset_and_size = offset_stack.next()
 
     def rename(self, rel_from, rel_to):
-        self._call('rename', 
+        self._call('rename',
                    self._remote_path(rel_from),
                    self._remote_path(rel_to))
 
     def move(self, rel_from, rel_to):
-        self._call('move', 
+        self._call('move',
                    self._remote_path(rel_from),
                    self._remote_path(rel_to))
 
     def rmdir(self, relpath):
         resp = self._call('rmdir', self._remote_path(relpath))
 
-    def _call(self, method, *args):
-        resp = self._client._call(method, *args)
-        self._translate_error(resp)
-
     def _translate_error(self, resp, orig_path=None):
         """Raise an exception from a response"""
         if resp is None:
@@ -877,20 +1233,14 @@
         else:
             raise errors.SmartProtocolError('unexpected smart server error: %r' % (resp,))
 
-    def _send_tuple(self, args):
-        self._client._send_tuple(args)
-
-    def _recv_tuple(self):
-        return self._client._recv_tuple()
-
     def disconnect(self):
-        self._client.disconnect()
+        self._medium.disconnect()
 
     def delete_tree(self, relpath):
         raise errors.TransportNotPossible('readonly transport')
 
     def stat(self, relpath):
-        resp = self._client._call('stat', self._remote_path(relpath))
+        resp = self._call2('stat', self._remote_path(relpath))
         if resp[0] == 'stat':
             return SmartStat(int(resp[1]), int(resp[2], 8))
         else:
@@ -913,148 +1263,510 @@
         return True
 
     def list_dir(self, relpath):
-        resp = self._client._call('list_dir',
-                                  self._remote_path(relpath))
+        resp = self._call2('list_dir', self._remote_path(relpath))
         if resp[0] == 'names':
             return [name.encode('ascii') for name in resp[1:]]
         else:
             self._translate_error(resp)
 
     def iter_files_recursive(self):
-        resp = self._client._call('iter_files_recursive',
-                                  self._remote_path(''))
+        resp = self._call2('iter_files_recursive', self._remote_path(''))
         if resp[0] == 'names':
             return resp[1:]
         else:
             self._translate_error(resp)
 
 
-class SmartStreamClient(SmartProtocolBase):
-    """Connection to smart server over two streams"""
-
-    def __init__(self, connect_func):
-        self._connect_func = connect_func
-        self._connected = False
-
-    def __del__(self):
-        self.disconnect()
-
-    def _ensure_connection(self):
-        if not self._connected:
-            self._in, self._out = self._connect_func()
-            self._connected = True
-
-    def _send_tuple(self, args):
-        self._ensure_connection()
-        return self._write_and_flush(_encode_tuple(args))
-
-    def _send_bulk_data(self, body):
-        self._ensure_connection()
-        SmartProtocolBase._send_bulk_data(self, body)
-        
-    def _recv_bulk(self):
-        self._ensure_connection()
-        return SmartProtocolBase._recv_bulk(self)
+class SmartClientMediumRequest(object):
+    """A request on a SmartClientMedium.
+
+    Each request allows bytes to be provided to it via accept_bytes, and then
+    the response bytes to be read via read_bytes.
+
+    For instance:
+    request.accept_bytes('123')
+    request.finished_writing()
+    result = request.read_bytes(3)
+    request.finished_reading()
+
+    It is up to the individual SmartClientMedium whether multiple concurrent
+    requests can exist. See SmartClientMedium.get_request to obtain instances 
+    of SmartClientMediumRequest, and the concrete Medium you are using for 
+    details on concurrency and pipelining.
+    """
+
+    def __init__(self, medium):
+        """Construct a SmartClientMediumRequest for the medium medium."""
+        self._medium = medium
+        # we track state by constants - we may want to use the same
+        # pattern as BodyReader if it gets more complex.
+        # valid states are: "writing", "reading", "done"
+        self._state = "writing"
+
+    def accept_bytes(self, bytes):
+        """Accept bytes for inclusion in this request.
+
+        This method may not be be called after finished_writing() has been
+        called.  It depends upon the Medium whether or not the bytes will be
+        immediately transmitted. Message based Mediums will tend to buffer the
+        bytes until finished_writing() is called.
+
+        :param bytes: A bytestring.
+        """
+        if self._state != "writing":
+            raise errors.WritingCompleted(self)
+        self._accept_bytes(bytes)
+
+    def _accept_bytes(self, bytes):
+        """Helper for accept_bytes.
+
+        Accept_bytes checks the state of the request to determing if bytes
+        should be accepted. After that it hands off to _accept_bytes to do the
+        actual acceptance.
+        """
+        raise NotImplementedError(self._accept_bytes)
+
+    def finished_reading(self):
+        """Inform the request that all desired data has been read.
+
+        This will remove the request from the pipeline for its medium (if the
+        medium supports pipelining) and any further calls to methods on the
+        request will raise ReadingCompleted.
+        """
+        if self._state == "writing":
+            raise errors.WritingNotComplete(self)
+        if self._state != "reading":
+            raise errors.ReadingCompleted(self)
+        self._state = "done"
+        self._finished_reading()
+
+    def _finished_reading(self):
+        """Helper for finished_reading.
+
+        finished_reading checks the state of the request to determine if 
+        finished_reading is allowed, and if it is hands off to _finished_reading
+        to perform the action.
+        """
+        raise NotImplementedError(self._finished_reading)
+
+    def finished_writing(self):
+        """Finish the writing phase of this request.
+
+        This will flush all pending data for this request along the medium.
+        After calling finished_writing, you may not call accept_bytes anymore.
+        """
+        if self._state != "writing":
+            raise errors.WritingCompleted(self)
+        self._state = "reading"
+        self._finished_writing()
+
+    def _finished_writing(self):
+        """Helper for finished_writing.
+
+        finished_writing checks the state of the request to determine if 
+        finished_writing is allowed, and if it is hands off to _finished_writing
+        to perform the action.
+        """
+        raise NotImplementedError(self._finished_writing)
+
+    def read_bytes(self, count):
+        """Read bytes from this requests response.
+
+        This method will block and wait for count bytes to be read. It may not
+        be invoked until finished_writing() has been called - this is to ensure
+        a message-based approach to requests, for compatability with message
+        based mediums like HTTP.
+        """
+        if self._state == "writing":
+            raise errors.WritingNotComplete(self)
+        if self._state != "reading":
+            raise errors.ReadingCompleted(self)
+        return self._read_bytes(count)
+
+    def _read_bytes(self, count):
+        """Helper for read_bytes.
+
+        read_bytes checks the state of the request to determing if bytes
+        should be read. After that it hands off to _read_bytes to do the
+        actual read.
+        """
+        raise NotImplementedError(self._read_bytes)
+
+
+class SmartClientStreamMediumRequest(SmartClientMediumRequest):
+    """A SmartClientMediumRequest that works with an SmartClientStreamMedium."""
+
+    def __init__(self, medium):
+        SmartClientMediumRequest.__init__(self, medium)
+        # check that we are safe concurrency wise. If some streams start
+        # allowing concurrent requests - i.e. via multiplexing - then this
+        # assert should be moved to SmartClientStreamMedium.get_request,
+        # and the setting/unsetting of _current_request likewise moved into
+        # that class : but its unneeded overhead for now. RBC 20060922
+        if self._medium._current_request is not None:
+            raise errors.TooManyConcurrentRequests(self._medium)
+        self._medium._current_request = self
+
+    def _accept_bytes(self, bytes):
+        """See SmartClientMediumRequest._accept_bytes.
+        
+        This forwards to self._medium._accept_bytes because we are operating
+        on the mediums stream.
+        """
+        self._medium._accept_bytes(bytes)
+
+    def _finished_reading(self):
+        """See SmartClientMediumRequest._finished_reading.
+
+        This clears the _current_request on self._medium to allow a new 
+        request to be created.
+        """
+        assert self._medium._current_request is self
+        self._medium._current_request = None
+        
+    def _finished_writing(self):
+        """See SmartClientMediumRequest._finished_writing.
+
+        This invokes self._medium._flush to ensure all bytes are transmitted.
+        """
+        self._medium._flush()
+
+    def _read_bytes(self, count):
+        """See SmartClientMediumRequest._read_bytes.
+        
+        This forwards to self._medium._read_bytes because we are operating
+        on the mediums stream.
+        """
+        return self._medium._read_bytes(count)
+
+
+class SmartClientRequestProtocolOne(SmartProtocolBase):
+    """The client-side protocol for smart version 1."""
+
+    def __init__(self, request):
+        """Construct a SmartClientRequestProtocolOne.
+
+        :param request: A SmartClientMediumRequest to serialise onto and
+            deserialise from.
+        """
+        self._request = request
+        self._body_buffer = None
+
+    def call(self, *args):
+        bytes = _encode_tuple(args)
+        self._request.accept_bytes(bytes)
+        self._request.finished_writing()
+
+    def call_with_body_bytes(self, args, body):
+        """Make a remote call of args with body bytes 'body'.
+
+        After calling this, call read_response_tuple to find the result out.
+        """
+        bytes = _encode_tuple(args)
+        self._request.accept_bytes(bytes)
+        bytes = self._encode_bulk_data(body)
+        self._request.accept_bytes(bytes)
+        self._request.finished_writing()
+
+    def call_with_body_readv_array(self, args, body):
+        """Make a remote call with a readv array.
+
+        The body is encoded with one line per readv offset pair. The numbers in
+        each pair are separated by a comma, and no trailing \n is emitted.
+        """
+        bytes = _encode_tuple(args)
+        self._request.accept_bytes(bytes)
+        readv_bytes = self._serialise_offsets(body)
+        bytes = self._encode_bulk_data(readv_bytes)
+        self._request.accept_bytes(bytes)
+        self._request.finished_writing()
+
+    def cancel_read_body(self):
+        """After expecting a body, a response code may indicate one otherwise.
+
+        This method lets the domain client inform the protocol that no body
+        will be transmitted. This is a terminal method: after calling it the
+        protocol is not able to be used further.
+        """
+        self._request.finished_reading()
+
+    def read_response_tuple(self, expect_body=False):
+        """Read a response tuple from the wire.
+
+        This should only be called once.
+        """
+        result = self._recv_tuple()
+        if not expect_body:
+            self._request.finished_reading()
+        return result
+
+    def read_body_bytes(self, count=-1):
+        """Read bytes from the body, decoding into a byte stream.
+        
+        We read all bytes at once to ensure we've checked the trailer for 
+        errors, and then feed the buffer back as read_body_bytes is called.
+        """
+        if self._body_buffer is not None:
+            return self._body_buffer.read(count)
+        _body_decoder = LengthPrefixedBodyDecoder()
+
+        while not _body_decoder.finished_reading:
+            bytes_wanted = _body_decoder.next_read_size()
+            bytes = self._request.read_bytes(bytes_wanted)
+            _body_decoder.accept_bytes(bytes)
+        self._request.finished_reading()
+        self._body_buffer = StringIO(_body_decoder.read_pending_data())
+        # XXX: TODO check the trailer result.
+        return self._body_buffer.read(count)
 
     def _recv_tuple(self):
-        self._ensure_connection()
-        return SmartProtocolBase._recv_tuple(self)
-
-    def _recv_trailer(self):
-        self._ensure_connection()
-        return SmartProtocolBase._recv_trailer(self)
-
-    def disconnect(self):
-        """Close connection to the server"""
-        if self._connected:
-            self._out.close()
-            self._in.close()
-
-    def _call(self, *args):
-        self._send_tuple(args)
-        return self._recv_tuple()
-
-    def _call_with_upload(self, method, args, body):
-        """Call an rpc, supplying bulk upload data.
-
-        :param method: method name to call
-        :param args: parameter args tuple
-        :param body: upload body as a byte string
-        """
-        self._send_tuple((method,) + args)
-        self._send_bulk_data(body)
-        return self._recv_tuple()
+        """Receive a tuple from the medium request."""
+        line = ''
+        while not line or line[-1] != '\n':
+            # TODO: this is inefficient - but tuples are short.
+            new_char = self._request.read_bytes(1)
+            line += new_char
+            assert new_char != '', "end of file reading from server."
+        return _decode_tuple(line)
 
     def query_version(self):
         """Return protocol version number of the server."""
-        # XXX: should make sure it's empty
-        self._send_tuple(('hello',))
-        resp = self._recv_tuple()
+        self.call('hello')
+        resp = self.read_response_tuple()
         if resp == ('ok', '1'):
             return 1
         else:
             raise errors.SmartProtocolError("bad response %r" % (resp,))
 
 
-class SmartTCPTransport(SmartTransport):
-    """Connection to smart server over plain tcp"""
-
-    def __init__(self, url, clone_from=None):
-        super(SmartTCPTransport, self).__init__(url, clone_from)
-        try:
-            self._port = int(self._port)
-        except (ValueError, TypeError), e:
-            raise errors.InvalidURL(path=url, extra="invalid port %s" % self._port)
-        self._socket = None
-
-    def _connect_to_server(self):
+class SmartClientMedium(object):
+    """Smart client is a medium for sending smart protocol requests over."""
+
+    def disconnect(self):
+        """If this medium maintains a persistent connection, close it.
+        
+        The default implementation does nothing.
+        """
+        
+
+class SmartClientStreamMedium(SmartClientMedium):
+    """Stream based medium common class.
+
+    SmartClientStreamMediums operate on a stream. All subclasses use a common
+    SmartClientStreamMediumRequest for their requests, and should implement
+    _accept_bytes and _read_bytes to allow the request objects to send and
+    receive bytes.
+    """
+
+    def __init__(self):
+        self._current_request = None
+
+    def accept_bytes(self, bytes):
+        self._accept_bytes(bytes)
+
+    def __del__(self):
+        """The SmartClientStreamMedium knows how to close the stream when it is
+        finished with it.
+        """
+        self.disconnect()
+
+    def _flush(self):
+        """Flush the output stream.
+        
+        This method is used by the SmartClientStreamMediumRequest to ensure that
+        all data for a request is sent, to avoid long timeouts or deadlocks.
+        """
+        raise NotImplementedError(self._flush)
+
+    def get_request(self):
+        """See SmartClientMedium.get_request().
+
+        SmartClientStreamMedium always returns a SmartClientStreamMediumRequest
+        for get_request.
+        """
+        return SmartClientStreamMediumRequest(self)
+
+    def read_bytes(self, count):
+        return self._read_bytes(count)
+
+
+class SmartSimplePipesClientMedium(SmartClientStreamMedium):
+    """A client medium using simple pipes.
+    
+    This client does not manage the pipes: it assumes they will always be open.
+    """
+
+    def __init__(self, readable_pipe, writeable_pipe):
+        SmartClientStreamMedium.__init__(self)
+        self._readable_pipe = readable_pipe
+        self._writeable_pipe = writeable_pipe
+
+    def _accept_bytes(self, bytes):
+        """See SmartClientStreamMedium.accept_bytes."""
+        self._writeable_pipe.write(bytes)
+
+    def _flush(self):
+        """See SmartClientStreamMedium._flush()."""
+        self._writeable_pipe.flush()
+
+    def _read_bytes(self, count):
+        """See SmartClientStreamMedium._read_bytes."""
+        return self._readable_pipe.read(count)
+
+
+class SmartSSHClientMedium(SmartClientStreamMedium):
+    """A client medium using SSH."""
+    
+    def __init__(self, host, port=None, username=None, password=None,
+            vendor=None):
+        """Creates a client that will connect on the first use.
+        
+        :param vendor: An optional override for the ssh vendor to use. See
+            bzrlib.transport.ssh for details on ssh vendors.
+        """
+        SmartClientStreamMedium.__init__(self)
+        self._connected = False
+        self._host = host
+        self._password = password
+        self._port = port
+        self._username = username
+        self._read_from = None
+        self._ssh_connection = None
+        self._vendor = vendor
+        self._write_to = None
+
+    def _accept_bytes(self, bytes):
+        """See SmartClientStreamMedium.accept_bytes."""
+        self._ensure_connection()
+        self._write_to.write(bytes)
+
+    def disconnect(self):
+        """See SmartClientMedium.disconnect()."""
+        if not self._connected:
+            return
+        self._read_from.close()
+        self._write_to.close()
+        self._ssh_connection.close()
+        self._connected = False
+
+    def _ensure_connection(self):
+        """Connect this medium if not already connected."""
+        if self._connected:
+            return
+        executable = os.environ.get('BZR_REMOTE_PATH', 'bzr')
+        if self._vendor is None:
+            vendor = ssh._get_ssh_vendor()
+        else:
+            vendor = self._vendor
+        self._ssh_connection = vendor.connect_ssh(self._username,
+                self._password, self._host, self._port,
+                command=[executable, 'serve', '--inet', '--directory=/',
+                         '--allow-writes'])
+        self._read_from, self._write_to = \
+            self._ssh_connection.get_filelike_channels()
+        self._connected = True
+
+    def _flush(self):
+        """See SmartClientStreamMedium._flush()."""
+        self._write_to.flush()
+
+    def _read_bytes(self, count):
+        """See SmartClientStreamMedium.read_bytes."""
+        if not self._connected:
+            raise errors.MediumNotConnected(self)
+        return self._read_from.read(count)
+
+
+class SmartTCPClientMedium(SmartClientStreamMedium):
+    """A client medium using TCP."""
+    
+    def __init__(self, host, port):
+        """Creates a client that will connect on the first use."""
+        SmartClientStreamMedium.__init__(self)
+        self._connected = False
+        self._host = host
+        self._port = port
+        self._socket = None
+
+    def _accept_bytes(self, bytes):
+        """See SmartClientMedium.accept_bytes."""
+        self._ensure_connection()
+        self._socket.sendall(bytes)
+
+    def disconnect(self):
+        """See SmartClientMedium.disconnect()."""
+        if not self._connected:
+            return
+        self._socket.close()
+        self._socket = None
+        self._connected = False
+
+    def _ensure_connection(self):
+        """Connect this medium if not already connected."""
+        if self._connected:
+            return
         self._socket = socket.socket()
         self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
         result = self._socket.connect_ex((self._host, int(self._port)))
         if result:
             raise errors.ConnectionError("failed to connect to %s:%d: %s" %
                     (self._host, self._port, os.strerror(result)))
-        # TODO: May be more efficient to just treat them as sockets
-        # throughout?  But what about pipes to ssh?...
-        to_server = self._socket.makefile('w')
-        from_server = self._socket.makefile('r')
-        return from_server, to_server
-
-    def disconnect(self):
-        super(SmartTCPTransport, self).disconnect()
-        # XXX: Is closing the socket as well as closing the files really
-        # necessary?
-        if self._socket is not None:
-            self._socket.close()
+        self._connected = True
+
+    def _flush(self):
+        """See SmartClientStreamMedium._flush().
+        
+        For TCP we do no flushing. We may want to turn off TCP_NODELAY and 
+        add a means to do a flush, but that can be done in the future.
+        """
+
+    def _read_bytes(self, count):
+        """See SmartClientMedium.read_bytes."""
+        if not self._connected:
+            raise errors.MediumNotConnected(self)
+        return self._socket.recv(count)
+
+
+class SmartTCPTransport(SmartTransport):
+    """Connection to smart server over plain tcp.
+    
+    This is essentially just a factory to get 'RemoteTransport(url,
+        SmartTCPClientMedium).
+    """
+
+    def __init__(self, url):
+        _scheme, _username, _password, _host, _port, _path = \
+            transport.split_url(url)
+        try:
+            _port = int(_port)
+        except (ValueError, TypeError), e:
+            raise errors.InvalidURL(path=url, extra="invalid port %s" % _port)
+        medium = SmartTCPClientMedium(_host, _port)
+        super(SmartTCPTransport, self).__init__(url, medium=medium)
 
 
 class SmartSSHTransport(SmartTransport):
-    """Connection to smart server over SSH."""
-
-    def __init__(self, url, clone_from=None):
-        # TODO: all this probably belongs in the parent class.
-        super(SmartSSHTransport, self).__init__(url, clone_from)
+    """Connection to smart server over SSH.
+
+    This is essentially just a factory to get 'RemoteTransport(url,
+        SmartSSHClientMedium).
+    """
+
+    def __init__(self, url):
+        _scheme, _username, _password, _host, _port, _path = \
+            transport.split_url(url)
         try:
-            if self._port is not None:
-                self._port = int(self._port)
+            if _port is not None:
+                _port = int(_port)
         except (ValueError, TypeError), e:
-            raise errors.InvalidURL(path=url, extra="invalid port %s" % self._port)
-
-    def _connect_to_server(self):
-        from bzrlib.transport import ssh
-        executable = os.environ.get('BZR_REMOTE_PATH', 'bzr')
-        vendor = ssh._get_ssh_vendor()
-        self._ssh_connection = vendor.connect_ssh(self._username,
-                self._password, self._host, self._port,
-                command=[executable, 'serve', '--inet', '--directory=/',
-                         '--allow-writes'])
-        return self._ssh_connection.get_filelike_channels()
-
-    def disconnect(self):
-        super(SmartSSHTransport, self).disconnect()
-        self._ssh_connection.close()
+            raise errors.InvalidURL(path=url, extra="invalid port %s" % 
+                _port)
+        medium = SmartSSHClientMedium(_host, _port, _username, _password)
+        super(SmartSSHTransport, self).__init__(url, medium=medium)
 
 
 def get_test_permutations():
-    """Return (transport, server) permutations for testing"""
+    """Return (transport, server) permutations for testing."""
+    ### We may need a little more test framework support to construct an
+    ### appropriate RemoteTransport in the future.
     return [(SmartTCPTransport, SmartTCPServer_for_testing)]

